import inspect
import typing
from typing import TYPE_CHECKING, TypeVar


# Inherits from type so that _WithRepr instances are types and can be used as
# e.g. Sequence[_WithRepr(...)]
class WithRepr(type):
    def __new__(mcs, obj, string):
        out = super().__new__(mcs, string, (), {})
        # prevent the custom typing repr from doing the wrong thing
        out.__module__ = "builtins"
        return out

    def __init__(cls, obj, string):
        cls.obj = obj
        cls.string = string

    def __repr__(cls):
        return cls.string

    def __call__(cls, *args, **kwargs):
        return cls.obj(*args, **kwargs)


_T = TypeVar("_T")


def doc_repr(obj: _T, string: str) -> _T:
    if TYPE_CHECKING:
        return obj
    else:
        if getattr(typing, "GENERATING_DOCUMENTATION", False):
            return WithRepr(obj, string)
        else:
            return obj


def doc_remove_args(*args):
    def doc_remove_args_impl(fn):
        if getattr(typing, "GENERATING_DOCUMENTATION", False):
            sig = inspect.signature(fn)
            new_params = []
            for param in sig.parameters.values():
                if param.name not in args:
                    new_params.append(param)
            sig = sig.replace(parameters=new_params)
            # Force-set it when doing `doc_remove_args(filter_jit(...))`.
            object.__setattr__(fn, "__signature__", sig)
        return fn

    return doc_remove_args_impl



import functools as ft
from typing import Any

import jax
import jax.core
import jax.interpreters.ad as ad
import jax.interpreters.batching as batching
import jax.interpreters.mlir as mlir
import jax.numpy as jnp
import jax.tree_util as jtu
from jaxtyping import PyTree

from .._filters import combine, is_array, is_array_like, partition
from .._tree import tree_equal


#
# filter primitives
# -----------------
# As with all filtering in Equinox, this is basically just about putting a
# nicer interface on existing JAX operations; in this case creating custom
# primitives. The inputs and outputs to the primitive can be arbitrary.
#


_like_sentinel = object()
_dummy_none = object()
_missing_dynamic = object()


# Not a PyTree
class _WrappedPrimal:
    def __init__(self, value):
        self.value = value


def _wrap_undefined(x):
    if isinstance(x, ad.UndefinedPrimal):
        return _WrappedPrimal(x)
    else:
        return x


def _unwrap_undefined(x, aval=False):
    if isinstance(x, _WrappedPrimal):
        if aval:
            return x.value.aval
        else:
            return x.value
    else:
        return x


def _is_array_like_internal(x):
    assert type(x) is not ad.UndefinedPrimal
    # Not `type(x) in (...)` as that doesn't handle stuff like ConcreteArrays.
    return is_array(x) or isinstance(x, (jax.core.ShapedArray, _WrappedPrimal))


def _zero_from_primal(p):
    assert type(p) is not ad.UndefinedPrimal
    aval = jax.core.get_aval(p)
    if hasattr(aval, "to_tangent_aval"):
        # JAX >=0.4.34
        aval = aval.to_tangent_aval()  # pyright: ignore
    else:
        # earlier JAX
        aval = aval.at_least_vspace()
    return ad.Zero(aval)


def _combine(dynamic, static):
    iter_dynamic = iter(dynamic)
    out = [next(iter_dynamic) if x is _missing_dynamic else x for x in static]
    assert next(iter_dynamic, None) is None
    return out


def _is_none(x):
    return x is None


def _replace_none(x):
    return _dummy_none if x is None else x


def _get_second(x, y):
    return None if x is None else y


def _make_spec(x, y):
    if y is None:
        return _is_array_like_internal(x)
    else:
        assert x is not None
        return True


class Flatten:
    __slots__ = ("treedef_out", "static_out")

    def called(self):
        return hasattr(self, "treedef_out")

    def get(self):
        return self.treedef_out, self.static_out

    def __call__(self, out, like=_like_sentinel):
        if like is _like_sentinel:
            dynamic_out, static_out = partition(out, _is_array_like_internal)
            flat_out, treedef_out = jtu.tree_flatten(dynamic_out)
            try:
                treedef_out_old = self.treedef_out
                static_out_old = self.static_out
            except AttributeError:
                self.treedef_out = treedef_out
                self.static_out = static_out
            else:
                assert treedef_out_old == treedef_out
                assert tree_equal(static_out_old, static_out)
            return flat_out
        else:
            assert jtu.tree_structure(out, is_leaf=_is_none) == jtu.tree_structure(
                like, is_leaf=_is_none
            )
            spec = jtu.tree_map(_make_spec, out, like, is_leaf=_is_none)
            dynamic_out, static_out = partition(out, spec, is_leaf=_is_none)
            flat_out, treedef_out = jtu.tree_flatten(dynamic_out)
            try:
                treedef_out_old = self.treedef_out
                static_out_old = self.static_out
            except AttributeError:
                self.treedef_out = treedef_out
                self.static_out = static_out
            else:
                assert treedef_out_old == treedef_out
                assert tree_equal(static_out_old, static_out)
            like = jtu.tree_map(_replace_none, like, is_leaf=_is_none)
            like = jtu.tree_map(_get_second, dynamic_out, like, is_leaf=_is_none)
            flat_like, treedef_like = jtu.tree_flatten(like)
            flat_like = [None if x is _dummy_none else x for x in flat_like]
            assert treedef_like == treedef_out
            assert len(flat_out) == len(flat_like)
            return flat_out, flat_like


def filter_primitive_def(rule):
    """For wrapping def_impl and def_abstract_eval.

    These can now take arbitrary inputs and outputs.
    """

    def _wrapper(*dynamic, treedef, static, flatten):
        args = jtu.tree_unflatten(treedef, _combine(dynamic, static))
        out = rule(*args)
        return flatten(out)

    return _wrapper


def filter_primitive_jvp(rule):
    """
    The input tangents (to the wrapped rule) will be a PyTree with the same
    structure as the input primals. `None` indicates symbolic zero tangents,
    in particular for non-JAX-array-likes.

    The output tangents are expected to match the output primals, necessarily
    with `None` for all non-JAX-array-likes.
    """

    def _wrapper(primals, tangents, *, treedef, static, flatten):
        tangents = [None if type(t) is ad.Zero else t for t in tangents]
        tangents_static = [x if x is _missing_dynamic else None for x in static]
        primals = jtu.tree_unflatten(treedef, _combine(primals, static))
        tangents = jtu.tree_unflatten(treedef, _combine(tangents, tangents_static))
        primals_out, tangents_out = rule(primals, tangents)
        flat_primals_out, flat_tangents_out = flatten(primals_out, tangents_out)
        flat_tangents_out = [
            _zero_from_primal(p) if t is None else t
            for p, t in zip(flat_primals_out, flat_tangents_out)
        ]
        return flat_primals_out, flat_tangents_out

    return _wrapper


_sentinel: Any = object()


def filter_primitive_transpose(rule=_sentinel, *, materialise_zeros=False):
    """
    The `inputs` to the transpose rule are a PyTree like the primal
    inputs, with `UndefinedPrimal`s where appropriate.

    The `cts_out` passed to the transpose rule are a PyTree like the
    primal output, with `None` for symbolic zero cotangents, in particular
    for non-JAX-array-likes.

    The output from the rule should be a PyTree like the primal input.
    All leaves which were non-JAX-array-like, or which should have zero
    cotangent, should have cotangent `None`.
    """
    if rule is _sentinel:
        return ft.partial(
            filter_primitive_transpose, materialise_zeros=materialise_zeros
        )

    def _wrapper(cts_out, *dynamic, treedef, static, flatten):
        treedef_out, _ = flatten.get()
        if materialise_zeros:
            cts_out = [ad.instantiate_zeros(ct) for ct in cts_out]
        else:
            cts_out = [None if type(ct) is ad.Zero else ct for ct in cts_out]
        cts_out = jtu.tree_unflatten(treedef_out, cts_out)
        wrapped_dynamic = [_wrap_undefined(x) for x in dynamic]
        wrapped_flat = _combine(wrapped_dynamic, static)
        wrapped_inputs = jtu.tree_unflatten(treedef, wrapped_flat)
        inputs = jtu.tree_map(_unwrap_undefined, wrapped_inputs)
        cts = rule(inputs, cts_out)
        flat_inputs, flat_cts = Flatten()(wrapped_inputs, cts)
        flat_inputs = [_unwrap_undefined(p, aval=True) for p in flat_inputs]
        flat_cts = [
            _zero_from_primal(p) if ct is None else ct
            for p, ct in zip(flat_inputs, flat_cts)
        ]
        assert len(dynamic) == len(flat_cts)
        return flat_cts

    return _wrapper


def filter_primitive_batching(rule):
    """
    The input batch axes (to the wrapped rule) will be a PyTree with the same
    structure as the input primals, with `None` for all non-JAX-arrays.

    The output batch axes are expected to match the output primals, with `None`
    for all non-JAX-arrays.
    """

    def _wrapper(dynamic, batch_axes, *, treedef, static, flatten):
        flat = _combine(dynamic, static)
        inputs = jtu.tree_unflatten(treedef, flat)
        batch_axes = [None if b is batching.not_mapped else b for b in batch_axes]
        batch_axes_static = [x if x is _missing_dynamic else None for x in static]
        batch_axes = _combine(batch_axes, batch_axes_static)
        batch_axes = jtu.tree_unflatten(treedef, batch_axes)
        out, batch_axes = rule(inputs, batch_axes)
        flat_out, flat_batch_axes = flatten(out, batch_axes)
        flat_batch_axes = [
            batching.not_mapped if b is None else b for b in flat_batch_axes
        ]
        return flat_out, flat_batch_axes

    return _wrapper


def filter_primitive_bind(prim: jax.core.Primitive, *args) -> PyTree:
    """Calls a primitive that has had its rules defined using the filter
    functions above.
    """
    assert prim.multiple_results
    # If `args` contains a Jaxpr or ClosedJaxpr in its leaves, then it ends up as a
    # member of the `static` tuple. This is important to ensure that jaxpr-rewriting
    # passes are able to find it.
    # (E.g. if `eqx.filter_closure_convert(...)` is an argument and we apply
    # `jax.core.jaxprs_in_params`.)
    flat, treedef = jtu.tree_flatten(args)
    dynamic = [x for x in flat if is_array(x)]
    static = tuple(_missing_dynamic if is_array(x) else x for x in flat)
    flatten = Flatten()
    flat_out = prim.bind(*dynamic, treedef=treedef, static=static, flatten=flatten)
    treedef_out, static_out = flatten.get()
    return combine(jtu.tree_unflatten(treedef_out, flat_out), static_out)


# Useful helper for JVP rules of higher-order primitives.
def materialise_zeros(primal, tangent, allow_struct=False):
    arraylike = is_array_like(primal)
    if allow_struct:
        arraylike = arraylike or isinstance(primal, jax.ShapeDtypeStruct)
    if tangent is None and arraylike:
        tangent = _zero_from_primal(primal)
        return ad.instantiate_zeros(tangent)
    else:
        return tangent


#
# vprim
# -----
# This allows for creating a primitive without needing to specify its batching rule.
# This is instead automatically obtained by vmap'ing its other rules (assuming that
# its rules are implemented using JAX operations).
#

_vprim_impl_registry = {}
_vprim_abstract_eval_registry = {}
_vprim_jvp_registry = {}
_vprim_transpose_registry = {}


def create_vprim(name: str, impl, abstract_eval, jvp, transpose):
    prim = jax.core.Primitive(name)
    prim.multiple_results = True

    def batch_rule(axis_size, axis_name, trace_type, inputs, batch_axes, **params):
        del trace_type
        # delegates batching to `_vprim_p`
        out = _vprim_p.bind(
            *inputs,
            prim=prim,
            __axis_size=axis_size,
            __axis_name=axis_name,
            __batch_axes=batch_axes,
            params=params,
        )
        batch_axes_out = jtu.tree_map(lambda _: 0, out)
        return out, batch_axes_out

    prim.def_impl(impl)
    prim.def_abstract_eval(abstract_eval)
    ad.primitive_jvps[prim] = jvp
    ad.primitive_transposes[prim] = transpose
    batching.axis_primitive_batchers[prim] = batch_rule
    mlir.register_lowering(prim, mlir.lower_fun(impl, multiple_results=True))
    _vprim_impl_registry[prim] = impl
    _vprim_abstract_eval_registry[prim] = abstract_eval
    _vprim_jvp_registry[prim] = jvp
    _vprim_transpose_registry[prim] = transpose
    return prim


def _vprim_impl(*inputs, prim, __axis_size, __axis_name, __batch_axes, params):
    impl = ft.partial(_vprim_impl_registry[prim], **params)
    impl = jax.vmap(
        impl, in_axes=__batch_axes, axis_size=__axis_size, axis_name=__axis_name
    )
    return impl(*inputs)


def _vprim_abstract_eval(*inputs, prim, __axis_size, __axis_name, __batch_axes, params):
    assert len(inputs) == len(__batch_axes)
    inputs = [
        jax.core.mapped_aval(__axis_size, b, x) for x, b in zip(inputs, __batch_axes)
    ]
    abstract_eval = _vprim_abstract_eval_registry[prim]
    outs = abstract_eval(*inputs, **params)
    outs = [jax.core.unmapped_aval(__axis_size, __axis_name, 0, x) for x in outs]
    return outs


def _resolve_zeros_t(tangent, batch_axis):
    if type(tangent) is ad.Zero and isinstance(batch_axis, int):
        aval = tangent.aval
        # Also accepts ConcreteArrays
        if not isinstance(aval, jax.core.ShapedArray):
            raise NotImplementedError(
                "vprim only currently supports ShapedArrays for symbolic zeros"
            )
        shape = aval.shape[:batch_axis] + aval.shape[batch_axis + 1 :]
        return ad.Zero(jax.core.ShapedArray(shape, aval.dtype))
    else:
        return tangent


def _resolve_zeros_b(tangent, batch_axis):
    if type(tangent) is ad.Zero:
        return None
    else:
        return batch_axis


def _vprim_jvp(
    primals, tangents, *, prim, __axis_size, __axis_name, __batch_axes, params
):
    assert len(primals) == len(__batch_axes)
    assert len(tangents) == len(__batch_axes)
    tangents = [_resolve_zeros_t(t, b) for t, b in zip(tangents, __batch_axes)]
    batch_axes_t = [_resolve_zeros_b(t, b) for t, b in zip(tangents, __batch_axes)]
    jvp = ft.partial(_vprim_jvp_registry[prim], **params)
    jvp = jax.vmap(
        jvp,
        in_axes=(__batch_axes, batch_axes_t),
        axis_size=__axis_size,
        axis_name=__axis_name,
    )
    return jvp(primals, tangents)


def _resolve_undefined_i(input, batch_axis):
    if type(input) is ad.UndefinedPrimal and isinstance(batch_axis, int):
        aval = input.aval
        # Also accepts ConcreteArrays
        if not isinstance(aval, jax.core.ShapedArray):
            raise NotImplementedError(
                "vprim only currently supports ShapedArrays for undefined primals"
            )
        shape = aval.shape[:batch_axis] + aval.shape[batch_axis + 1 :]
        return ad.UndefinedPrimal(jax.core.ShapedArray(shape, aval.dtype))
    else:
        return input


def _resolve_undefined_b(input, batch_axis):
    if type(input) is ad.UndefinedPrimal:
        return None
    else:
        return batch_axis


def _vprim_transpose(
    cts, *inputs, prim, __axis_size, __axis_name, __batch_axes, params
):
    mapped_inputs = [_resolve_undefined_i(i, b) for i, b in zip(inputs, __batch_axes)]
    batch_axes = [_resolve_undefined_b(i, b) for i, b in zip(inputs, __batch_axes)]

    def _transpose(*_inputs):
        _outputs = _vprim_transpose_registry[prim](*_inputs, **params)
        # `Zero` is not a JAX type -- it's an internal AD thing -- so we shouldn't pass
        # it across the `vmap` boundary. In particular JAX won't apply the out batch
        # axis to it.
        # JAX allows for returning `None` to indicate no cotangent, so we use that
        # instead, which is compatible with both `vmap` and `out_axes`.
        return tuple(None if type(o) is ad.Zero else o for o in _outputs)

    transpose = jax.vmap(
        _transpose,
        in_axes=(0, *batch_axes),
        out_axes=__batch_axes,
        axis_size=__axis_size,
        axis_name=__axis_name,
    )
    if prim.multiple_results:
        cts = tuple(None if type(c) is ad.Zero else c for c in cts)
    else:
        cts = None if type(cts) is ad.Zero else cts
    outputs = transpose(cts, *mapped_inputs)
    assert len(inputs) == len(outputs)
    for i, o in zip(inputs, outputs):
        if o is not None:
            # Can't have cotangents on defined variables I think? The point of an
            # `UndefinedPrimal` is to declare what you want cotangents with respect to.
            assert type(i) is ad.UndefinedPrimal
            # We've filtered out all other avals above, with a `NotImplementedError` if
            # required.
            assert isinstance(i.aval, jax.core.ShapedArray)
            assert i.aval.shape == jnp.shape(o)
    return outputs


# _vprim_p is itself a vprim!
_vprim_p = create_vprim(
    "vprim", _vprim_impl, _vprim_abstract_eval, _vprim_jvp, _vprim_transpose
)



import os
import warnings
from typing import Literal


EQX_ON_ERROR: Literal["raise", "breakpoint", "nan"] = os.environ.get(  # pyright: ignore
    "EQX_ON_ERROR", "raise"
)
if EQX_ON_ERROR not in ("raise", "breakpoint", "nan"):
    raise ValueError(
        "Unrecognised value for `EQX_ON_ERROR`. Valid values are `EQX_ON_ERROR=raise`, "
        "`EQX_ON_ERROR=breakpoint`, and `EQX_ON_ERROR=nan`."
    )
if EQX_ON_ERROR == "breakpoint":
    warnings.warn(
        "The environment variable `EQX_ON_ERROR=breakpoint` is currently set. Note "
        "that this should only be used for debugging, as it slows down runtime speed."
    )


EQX_ON_ERROR_BREAKPOINT_FRAMES = os.environ.get("EQX_ON_ERROR_BREAKPOINT_FRAMES", None)
if EQX_ON_ERROR_BREAKPOINT_FRAMES is None:
    EQX_ON_ERROR_BREAKPOINT_FRAMES = 1
else:
    EQX_ON_ERROR_BREAKPOINT_FRAMES = int(EQX_ON_ERROR_BREAKPOINT_FRAMES)

try:
    EQX_GETKEY_SEED = int(os.environ["EQX_GETKEY_SEED"])
except KeyError:
    EQX_GETKEY_SEED = None



import atexit
import functools as ft
import inspect
import logging
import warnings
from collections.abc import Callable
from typing import Any, Literal, overload, TypeVar
from typing_extensions import ParamSpec

import jax
import jax._src.dispatch
import jax._src.traceback_util as traceback_util
import jax.core
import jax.errors
import jax.numpy as jnp
from jaxtyping import PyTree

from ._compile_utils import (
    compile_cache,
    get_fn_names,
    hashable_combine,
    hashable_filter,
    hashable_partition,
    Lowered,
)
from ._custom_types import sentinel
from ._deprecate import deprecated_0_10
from ._doc_utils import doc_remove_args
from ._filters import combine, is_array, partition
from ._misc import currently_jitting
from ._module import field, Module, module_update_wrapper, Partial, Static


traceback_util.register_exclusion(__file__)


_P = ParamSpec("_P")
_T = TypeVar("_T")


@compile_cache
def _filter_jit_cache(fun_names, jitkwargs):
    def fun_wrapped(dynamic_donate, dynamic_nodonate, static):
        dynamic_dict = dict(**dynamic_donate, **dynamic_nodonate)
        dynamic_fun = dynamic_dict.pop("fun")
        dynamic_first = dynamic_dict.pop("first")
        dynamic_rest = dynamic_dict.pop("rest")
        static_fun, static_first, static_rest = static
        fun = hashable_combine(dynamic_fun, static_fun)
        first_arg = hashable_combine(dynamic_first, static_first)
        rest_args, kwargs = hashable_combine(dynamic_rest, static_rest)
        assert type(rest_args) is tuple
        *args, dummy_arg = (first_arg,) + rest_args
        assert dummy_arg is None
        out = fun(*args, **kwargs)
        dynamic_out, static_out = partition(out, is_array)
        marker = jnp.array(0)
        return marker, dynamic_out, Static(static_out)

    fun_name, fun_qualname = fun_names
    fun_wrapped.__name__ = fun_name
    fun_wrapped.__qualname__ = fun_qualname
    return jax.jit(fun_wrapped, donate_argnums=0, static_argnums=2, **jitkwargs)


def _bind(signature, args, kwargs):
    bound = signature.bind(*args, **kwargs)
    args = bound.args
    kwargs = bound.kwargs
    return args, kwargs


def _preprocess(info, args, kwargs, return_static: bool = False):
    signature, dynamic_fun, static_fun, donate_first, donate_rest = info
    args, kwargs = _bind(signature, args, kwargs)
    # add dummy to avoid special casing `len(args) == 0`.
    args = args + (None,)
    first_arg = args[0]
    rest_args = args[1:]
    if return_static:
        dynamic_first, static_first = hashable_partition(first_arg, is_array)
        dynamic_rest, static_rest = hashable_partition((rest_args, kwargs), is_array)
    else:
        dynamic_first = hashable_filter(first_arg, is_array)
        dynamic_rest = hashable_filter((rest_args, kwargs), is_array)
    dynamic_donate = dict()
    dynamic_nodonate = dict()
    if donate_first:
        dynamic_donate["first"] = dynamic_first
    else:
        dynamic_nodonate["first"] = dynamic_first
    if donate_rest:
        dynamic_donate["fun"] = dynamic_fun
        dynamic_donate["rest"] = dynamic_rest
    else:
        dynamic_nodonate["fun"] = dynamic_fun
        dynamic_nodonate["rest"] = dynamic_rest

    if return_static:
        static = (static_fun, static_first, static_rest)  # pyright: ignore
        return dynamic_donate, dynamic_nodonate, static
    else:
        return dynamic_donate, dynamic_nodonate


def _postprocess(out):
    _, dynamic_out, static_out = out
    return combine(dynamic_out, static_out.value)


try:
    # Added in JAX 0.4.34.
    JaxRuntimeError = jax.errors.JaxRuntimeError  # pyright: ignore
except AttributeError:
    try:
        # Forward compatibility in case they ever decide to fix the capitalization.
        JaxRuntimeError = jax.errors.JAXRuntimeError  # pyright: ignore
    except AttributeError:
        # Not public API, so wrap in a try-except for forward compatibility.
        try:
            JaxRuntimeError = jax.lib.xla_extension.XlaRuntimeError  # pyright: ignore
        except Exception:
            # Unused dummy
            class JaxRuntimeError(Exception):
                pass


try:
    wait_for_tokens = jax._src.dispatch.wait_for_tokens
except AttributeError:
    pass  # forward compatibility
else:
    # Fix for https://github.com/patrick-kidger/diffrax/issues/506
    def wait_for_tokens2():
        try:
            wait_for_tokens()
        except JaxRuntimeError:
            pass

    atexit.unregister(wait_for_tokens)
    atexit.register(wait_for_tokens2)


# This is the class we use to raise runtime errors from `eqx.error_if`.
class EquinoxRuntimeError(RuntimeError):
    pass


# Magic value that means error messages are displayed as `{__qualname__}: ...` rather
# than `{__module__}.{__qualname__}`. (At least, I checked the default Python
# interpreter, the default Python REPL, ptpython, ipython, pdb, and ipdb.)
EquinoxRuntimeError.__module__ = "builtins"
# Note that we don't also override `__name__` or `__qualname__`. Suppressing the
# `equinox._jit` module bit is useful for readability, but we don't want to go so far as
# deleting the name altogether. (Or even e.g. setting it to the 'Above is the stack...'
# first section of our error message below!) The reason is that whilst that gives a
# nicer displayed error in default Python, it doesn't necessarily do as well with other
# tools, e.g. debuggers. So what we have here is a compromise.


last_msg = None
last_stack = None


_on_error_msg = """Above is the stack outside of JIT. Below is the stack inside of JIT:
{stack}
equinox.EquinoxRuntimeError: {msg}

-------------------

An error occurred during the runtime of your JAX program.

1) Setting the environment variable `EQX_ON_ERROR=breakpoint` is usually the most useful
way to debug such errors. This can be interacted with using most of the usual commands
for the Python debugger: `u` and `d` to move up and down frames, the name of a variable
to print its value, etc.

2) You may also like to try setting `JAX_DISABLE_JIT=1`. This will mean that you can
(mostly) inspect the state of your program as if it was normal Python.

3) See `https://docs.kidger.site/equinox/api/debug/` for more suggestions.
"""


class _FilterCallback(logging.Filterer):
    def filter(self, record: logging.LogRecord):
        return not (
            record.name == "jax._src.callback"
            and record.getMessage() == "jax.pure_callback failed"
        )


class _JitWrapper(Module):
    fn: str  # this attribute exists solely to give a nice repr
    _signature: inspect.Signature = field(static=True, repr=False)
    _dynamic_fun: PyTree = field(repr=False)
    _static_fun: Any = field(static=True, repr=False)
    _cached: Any = field(static=True, repr=False)
    filter_warning: bool = field(static=True)
    donate_first: bool = field(static=True)
    donate_rest: bool = field(static=True)

    @property
    def __wrapped__(self):
        return hashable_combine(self._dynamic_fun, self._static_fun)

    def _call(self, is_lower, args, kwargs):
        __tracebackhide__ = True
        # Used by our error messages when figuring out where to stop walking the stack.
        jitting = currently_jitting()
        if not jitting:
            __equinox_filter_jit__ = True  # noqa: F841
        info = (
            self._signature,
            self._dynamic_fun,
            self._static_fun,
            self.donate_first,
            self.donate_rest,
        )
        dynamic_donate, dynamic_nodonate, static = _preprocess(  # pyright: ignore
            info, args, kwargs, return_static=True
        )
        if is_lower:
            return Lowered(
                self._cached.lower(dynamic_donate, dynamic_nodonate, static),
                info,
                _preprocess,  # pyright: ignore
                _postprocess,  # pyright: ignore
            )
        else:
            filter = _FilterCallback()
            callback_logger = logging.getLogger("jax._src.callback")
            callback_logger.addFilter(filter)
            try:
                if self.filter_warning:
                    with warnings.catch_warnings():
                        warnings.filterwarnings(
                            "ignore", message="Some donated buffers were not usable*"
                        )
                        marker, _, _ = out = self._cached(
                            dynamic_donate, dynamic_nodonate, static
                        )
                else:
                    marker, _, _ = out = self._cached(
                        dynamic_donate, dynamic_nodonate, static
                    )
                if not jitting:
                    marker.block_until_ready()
            except JaxRuntimeError as e:
                # Catch Equinox's runtime errors, and re-raise them with actually useful
                # information. (By default XlaRuntimeError produces a lot of terrifying
                # but useless information.)
                if (
                    last_msg is not None
                    and last_stack is not None
                    and "_EquinoxRuntimeError: " in str(e)
                ):
                    # We check `last_msg` and `last_stack` just in case. I'm not sure
                    # what happens in distributed/multiprocess environments. Is the
                    # callback necessarily executed in the same interpreter as we are in
                    # here?
                    raise EquinoxRuntimeError(
                        _on_error_msg.format(msg=last_msg, stack=last_stack)
                    ) from None
                    # `from None` to hide the large but uninformative XlaRuntimeError.
                else:
                    raise
            finally:
                callback_logger.removeFilter(filter)
            return _postprocess(out)

    def __call__(self, /, *args, **kwargs):
        __tracebackhide__ = True
        try:
            return self._call(False, args, kwargs)
        except EquinoxRuntimeError as e:
            # Use a two-part try/except here and in `_call` to delete the
            # `raise EquinoxRuntimeError` line from the stack trace.
            e.__traceback__ = None
            raise

    def lower(self, /, *args, **kwargs) -> Lowered:
        return self._call(True, args, kwargs)

    def __get__(self, instance, owner):
        del owner
        if instance is None:
            return self
        return Partial(self, instance)


@overload
def filter_jit(
    *,
    donate: Literal[
        "all", "all-except-first", "warn", "warn-except-first", "none"
    ] = "none",
) -> Callable[[Callable[_P, _T]], Callable[_P, _T]]: ...


@overload
def filter_jit(
    fun: Callable[_P, _T],
    *,
    donate: Literal[
        "all", "all-except-first", "warn", "warn-except-first", "none"
    ] = "none",
) -> Callable[_P, _T]: ...


@doc_remove_args("jitkwargs")
def filter_jit(
    fun=sentinel,
    *,
    donate: Literal[
        "all", "all-except-first", "warn", "warn-except-first", "none"
    ] = "none",
    **jitkwargs,
):
    """An easier-to-use version of `jax.jit`. All JAX and NumPy arrays are traced, and
    all other types are held static.

    **Arguments:**

    - `fun` is a pure function to JIT compile.
    - `donate` indicates whether the buffers of JAX arrays are donated or not. It
        should either be:
        - `'all'`: donate all arrays and suppress all warnings about unused buffers;
        - `'all-except-first'`: donate all arrays except for those in the first
            argument, and suppress all warnings about unused buffers;
        - `'warn'`: as above, but don't suppress unused buffer warnings;
        - `'warn-except-first'`: as above, but don't suppress unused buffer warnings;
        - `'none'`: no buffer donation. (This the default.)

    **Returns:**

    The JIT'd version of `fun`.

    !!! example

        ```python
        # Basic behaviour
        @eqx.filter_jit
        def f(x, y):  # both args traced if arrays, static if non-arrays
            return x + y, x - y

        f(jnp.array(1), jnp.array(2))  # both args traced
        f(jnp.array(1), 2)  # first arg traced, second arg static
        f(1, 2)  # both args static
        ```

    !!! info

        Donating arguments allows their underlying memory to be used in the
        computation. This can produce speed and memory improvements, but means that you
        cannot use any donated arguments again, as their underlying memory has been
        overwritten. (JAX will throw an error if you try to.)

    !!! info

        If you want to trace Python `bool`/`int`/`float`/`complex` as well then you
        can do this by wrapping them into a JAX array: `jnp.asarray(x)`.

        If you want to donate only some arguments then this can be done by setting
        `filter_jit(donate="all-except-first")` and then passing all arguments that you
        don't want to donate through the first argument. (Packing multiple values into
        a tuple if necessary.)
    """

    if fun is sentinel:
        return ft.partial(filter_jit, donate=donate, **jitkwargs)

    deprecated_0_10(jitkwargs, "default")
    deprecated_0_10(jitkwargs, "fn")
    deprecated_0_10(jitkwargs, "args")
    deprecated_0_10(jitkwargs, "kwargs")
    deprecated_0_10(jitkwargs, "out")
    if any(
        x in jitkwargs for x in ("static_argnums", "static_argnames", "donate_argnums")
    ):
        raise ValueError(
            "`jitkwargs` cannot contain 'static_argnums', 'static_argnames' or "
            "'donate_argnums'"
        )
    signature = inspect.signature(fun)

    if donate == "all":
        filter_warning = True
        donate_first = True
        donate_rest = True
    elif donate == "all-except-first":
        filter_warning = True
        donate_first = False
        donate_rest = True
    elif donate == "warn":
        filter_warning = False
        donate_first = True
        donate_rest = True
    elif donate == "warn-except-first":
        filter_warning = False
        donate_first = False
        donate_rest = True
    elif donate == "none":
        filter_warning = False
        donate_first = False
        donate_rest = False
    else:
        raise ValueError(
            "`filter_jit(..., donate=...)` must be one of 'all', 'all-except-first', "
            "'warn', 'warn-except-first', or 'none'."
        )

    _, name = get_fn_names(fun)
    dynamic_fun, static_fun = hashable_partition(fun, is_array)
    cached = _filter_jit_cache(fun, jitkwargs)

    jit_wrapper = _JitWrapper(
        fn=name,
        _signature=signature,
        _dynamic_fun=dynamic_fun,
        _static_fun=static_fun,
        _cached=cached,
        filter_warning=filter_warning,
        donate_first=donate_first,
        donate_rest=donate_rest,
    )
    return module_update_wrapper(jit_wrapper)



import functools as ft
import warnings
from collections.abc import Sequence
from typing import Optional, overload, Union

import jax
import jax.numpy as jnp
from jaxtyping import Array, Float, PRNGKeyArray

from .._custom_types import sentinel
from .._misc import default_floating_dtype, left_broadcast_to
from .._module import field, Module
from ._stateful import State


class LayerNorm(Module, strict=True):
    r"""
    Computes a mean and standard deviation over the whole input array, and uses these
    to normalise the whole array. Optionally applies an elementwise affine
    transformation afterwards.

    Given an input array $x$, this layer computes

    $$\frac{x - \mathbb{E}[x]}{\sqrt{\text{Var}[x] + \varepsilon}} * \gamma + \beta$$

    where $\gamma$, $\beta$ have the same shape as $x$ if `elementwise_affine=True`,
    and $\gamma = 1$, $\beta = 0$ if `elementwise_affine=False`.

    ??? cite
        [Layer Normalization](https://arxiv.org/abs/1607.06450)

        ```bibtex
        @article{ba2016layer,
            author={Jimmy Lei Ba, Jamie Ryan Kriso, Geoffrey E. Hinton},
            title={Layer Normalization},
            year={2016},
            journal={arXiv:1607.06450},
        }
        ```

    !!! faq "FAQ"

        If you need to normalise over only some input dimensions, then this can be
        achieved by vmap'ing. For example the following will compute statistics over
        every dimension *except* the first:
        ```python
        layer = LayerNorm(...)
        array = jax.vmap(layer)(array)
        ```

    """

    shape: tuple[int, ...] = field(static=True)
    eps: float = field(static=True)
    use_weight: bool = field(static=True)
    use_bias: bool = field(static=True)
    weight: Optional[Float[Array, "*shape"]]
    bias: Optional[Float[Array, "*shape"]]

    def __init__(
        self,
        shape: Union[int, Sequence[int]],
        eps: float = 1e-5,
        use_weight: bool = True,
        use_bias: bool = True,
        dtype=None,
        *,
        elementwise_affine: Optional[bool] = None,
    ):
        """**Arguments:**

        - `shape`: Shape of the input.
        - `eps`: Value added to denominator for numerical stability.
        - `use_weight`: Whether the module has learnable affine weights.
        - `use_bias`: Whether the module has learnable affine biases.
        - `dtype`: The dtype to use for the weight and the bias in this layer if
            `use_weight` or `use_bias` is set to `True`.
            Defaults to either `jax.numpy.float32` or `jax.numpy.float64` depending
            on whether JAX is in 64-bit mode.
        - `elementwise_affine`: Deprecated alternative to `use_weight` and `use_bias`.
        """
        if isinstance(shape, int):
            shape = (shape,)
        else:
            shape = tuple(shape)
        self.shape = shape
        self.eps = eps
        if elementwise_affine is not None:
            use_weight = elementwise_affine
            use_bias = elementwise_affine
            warnings.warn(
                "LayerNorm(elementwise_affine=...) is deprecated "
                "in favour of LayerNorm(use_weight=...) and LayerNorm(use_bias=...)"
            )
        self.use_weight = use_weight
        self.use_bias = use_bias
        self.weight = jnp.ones(shape, dtype=dtype) if use_weight else None
        self.bias = jnp.zeros(shape, dtype=dtype) if use_bias else None

    @overload
    def __call__(self, x: Array, *, key: Optional[PRNGKeyArray] = None) -> Array: ...

    @overload
    def __call__(
        self, x: Array, state: State, *, key: Optional[PRNGKeyArray] = None
    ) -> tuple[Array, State]: ...

    @jax.named_scope("eqx.nn.LayerNorm")
    def __call__(
        self,
        x: Float[Array, "*shape"],
        state: State = sentinel,
        *,
        key: Optional[PRNGKeyArray] = None,
    ) -> Union[Array, tuple[Array, State]]:
        """**Arguments:**

        - `x`: A JAX array, with the same shape as the `shape` passed to `__init__`.
        - `state`: Ignored; provided for interchangeability with the
            [`equinox.nn.BatchNorm`][] API.
        - `key`: Ignored; provided for compatibility with the rest of the Equinox API.
            (Keyword only argument.)

        **Returns:**

        The output is a JAX array of the same shape as `x`.

        If `state` is passed, then a 2-tuple of `(output, state)` is returned. The state
        is passed through unchanged. If `state` is not passed, then just the output is
        returned.
        """
        if x.shape != self.shape:
            raise ValueError(
                "`LayerNorm(shape)(x)` must satisfy the invariant `shape == x.shape`"
                f"Received `shape={self.shape} and `x.shape={x.shape}`. You might need "
                "to replace `layer_norm(x)` with `jax.vmap(layer_norm)(x)`.\n"
                "\n"
                "If this is a new error for you, it might be because this became "
                "stricter in Equinox v0.11.0. Previously all that was required is that "
                "`x.shape` ended with `shape`. However, this turned out to be a "
                "frequent source of bugs, so we made the check stricter!"
            )
        orig_dtype = x.dtype
        with jax.numpy_dtype_promotion("standard"):
            dtype = jnp.result_type(x.dtype, jnp.float32)

        x = x.astype(dtype)
        mean = jnp.mean(x, keepdims=True)
        variance = jnp.var(x, keepdims=True)
        variance = jnp.maximum(0.0, variance)
        inv = jax.lax.rsqrt(variance + self.eps)
        out = (x - mean) * inv
        if self.use_weight:
            out = self.weight.astype(dtype) * out  # pyright: ignore
        if self.use_bias:
            out = out + self.bias.astype(dtype)  # pyright: ignore
        if state is sentinel:
            return out.astype(orig_dtype)
        else:
            return out.astype(orig_dtype), state


class GroupNorm(Module, strict=True):
    r"""
    Splits the first dimension ("channels") into groups of fixed size. Computes a mean
    and standard deviation over the contents of each group, and uses these to normalise
    the group. Optionally applies a channel-wise affine transformation afterwards.

    Given an input array $x$ of shape `(channels, ...)`, this layer splits this up into
    `groups`-many arrays $x_i$ each of shape `(channels/groups, ...)`, and for each one
    computes

    $$\frac{x_i - \mathbb{E}[x_i]}{\sqrt{\text{Var}[x_i] + \varepsilon}} * \gamma_i + \beta_i$$

    where $\gamma_i$, $\beta_i$ have shape `(channels/groups,)` if
    `channelwise_affine=True`, and $\gamma = 1$, $\beta = 0$ if
    `channelwise_affine=False`.

    ??? cite
        [Group Normalization](https://arxiv.org/abs/1803.08494)

        ```bibtex
        @article{wu2018group,
            author={Yuxin Wu and Kaiming He},
            title={Group Normalization},
            year={2018},
            journal={arXiv:1803.08494},
        }
        ```
    """  # noqa: E501

    groups: int = field(static=True)
    channels: Optional[int] = field(static=True)
    eps: float = field(static=True)
    channelwise_affine: bool = field(static=True)
    weight: Optional[Array]
    bias: Optional[Array]

    def __init__(
        self,
        groups: int,
        channels: Optional[int] = None,
        eps: float = 1e-5,
        channelwise_affine: bool = True,
        dtype=None,
    ):
        """**Arguments:**

        - `groups`: The number of groups to split the input into.
        - `channels`: The number of input channels. May be left unspecified (e.g. just
            `None`) if `channelwise_affine=False`.
        - `eps`: Value added to denominator for numerical stability.
        - `channelwise_affine`: Whether the module has learnable affine parameters.
        - `dtype`: The dtype to use for the weight and the bias in this layer if
            `channelwise_affine` is set to `True`.
            Defaults to either `jax.numpy.float32` or `jax.numpy.float64` depending
            on whether JAX is in 64-bit mode.
        """
        if (channels is not None) and (channels % groups != 0):
            raise ValueError("The number of groups must divide the number of channels.")
        if (channels is None) and channelwise_affine:
            raise ValueError(
                "The number of channels should be specified if "
                "`channelwise_affine=True`"
            )
        self.groups = groups
        self.channels = channels
        self.eps = eps
        self.channelwise_affine = channelwise_affine
        self.weight = jnp.ones(channels, dtype=dtype) if channelwise_affine else None
        self.bias = jnp.zeros(channels, dtype=dtype) if channelwise_affine else None

    @overload
    def __call__(self, x: Array, *, key: Optional[PRNGKeyArray] = None) -> Array: ...

    @overload
    def __call__(
        self, x: Array, state: State, *, key: Optional[PRNGKeyArray] = None
    ) -> tuple[Array, State]: ...

    @jax.named_scope("eqx.nn.GroupNorm")
    def __call__(
        self, x: Array, state: State = sentinel, *, key: Optional[PRNGKeyArray] = None
    ) -> Union[Array, tuple[Array, State]]:
        """**Arguments:**

        - `x`: A JAX array of shape `(channels, ...)`.
        - `state`: Ignored; provided for interchangability with the
            [`equinox.nn.BatchNorm`][] API.
        - `key`: Ignored; provided for compatibility with the rest of the Equinox API.
            (Keyword only argument.)

        **Returns:**

        The output is a JAX array of shape `(channels, ...)`.

        If `state` is passed, then a 2-tuple of `(output, state)` is returned. The state
        is passed through unchanged. If `state` is not passed, then just the output is
        returned.
        """

        orig_dtype = x.dtype
        with jax.numpy_dtype_promotion("standard"):
            dtype = jnp.result_type(x.dtype, jnp.float32)

        x = x.astype(dtype)
        channels = x.shape[0]
        y = x.reshape(self.groups, channels // self.groups, *x.shape[1:])
        mean = jax.vmap(ft.partial(jnp.mean, keepdims=True))(y)
        variance = jax.vmap(ft.partial(jnp.var, keepdims=True))(y)
        variance = jnp.maximum(0.0, variance)
        inv = jax.lax.rsqrt(variance + self.eps)
        out = (y - mean) * inv
        out = out.reshape(x.shape)
        if self.channelwise_affine:
            weight = left_broadcast_to(self.weight, out.shape)  # pyright: ignore
            bias = left_broadcast_to(self.bias, out.shape)  # pyright: ignore
            out = weight.astype(dtype) * out + bias.astype(dtype)
        if state is sentinel:
            return out.astype(orig_dtype)
        else:
            return out.astype(orig_dtype), state


class RMSNorm(Module, strict=True):
    r"""
    A simplified version of LayerNorm which rescales the inputs, but does not center
    them. Optionally applies a learned reweighting of the transformed array afterward.

    Given an input array $x$, this layer computes

    $$\frac{x}{\sqrt{\varepsilon + \frac{1}{n}\Vert x \Vert^2_2}} \gamma + \beta$$

    where $\Vert x \Vert^2_2 = \sum_{i=1}^n x_i^2$, $n = \dim(x)$, and $\gamma$ is a
    learned array with the same shape as $x$ if `use_weight=True`, or
    $\gamma = 1$ if `use_weight=False`, as proposed in
    [this paper](https://browse.arxiv.org/abs/2307.14995). `\beta` is an optional bias
    term.

    ??? cite

        [Root Mean Square Layer Normalization](https://browse.arxiv.org/abs/1910.07467)

        ```bibtex
        @article{zhang2019root,
            title={Root Mean Square Layer Normalization},
            author={Biao Zhang and Rico Sennrich},
            year={2019},
            journal={arXiv:1910.07467}
        }
        ```
    """

    shape: tuple[int, ...] = field(static=True)
    eps: float = field(static=True)
    use_weight: bool = field(static=True)
    use_bias: bool = field(static=True)
    weight: Optional[Float[Array, "*shape"]]
    bias: Optional[Float[Array, "*shape"]]

    def __init__(
        self,
        shape: Union[int, Sequence[int]],
        eps: float = 1e-5,
        use_weight: bool = True,
        use_bias: bool = True,
        dtype=None,
    ):
        """**Arguments:**

        - `shape`: Shape of the input.
        - `eps`: Value added to denominator for numerical stability.
        - `use_weight`: Whether the module has learnable affine weights.
        - `use_bias`: Whether the module has learnable affine shift.
        - `dtype`: The dtype to use for the weight and the bias in this layer if
            `use_weight` or `use_bias` is set to `True`.
            Defaults to either `jax.numpy.float32` or `jax.numpy.float64` depending
            on whether JAX is in 64-bit mode.
        """
        dtype = default_floating_dtype() if dtype is None else dtype
        if isinstance(shape, int):
            shape = (shape,)
        else:
            shape = tuple(shape)
        self.shape = shape
        self.eps = eps
        self.use_weight = use_weight
        self.use_bias = use_bias
        self.weight = jnp.ones(shape, dtype=dtype) if use_weight else None
        self.bias = jnp.zeros(shape, dtype=dtype) if use_bias else None

    @overload
    def __call__(self, x: Array, *, key: Optional[PRNGKeyArray] = None) -> Array: ...

    @overload
    def __call__(
        self, x: Array, state: State, *, key: Optional[PRNGKeyArray] = None
    ) -> tuple[Array, State]: ...

    @jax.named_scope("eqx.nn.RMSNorm")
    def __call__(
        self,
        x: Float[Array, "*shape"],
        state: State = sentinel,
        *,
        key: Optional[PRNGKeyArray] = None,
    ) -> Union[Array, tuple[Array, State]]:
        """**Arguments:**

        - `x`: A JAX array, with the same shape as the `shape` passed to `__init__`.
        - `state`: Ignored; provided for interchangability with the
            [`equinox.nn.BatchNorm`][] API.
        - `key`: Ignored; provided for compatibility with the rest of the Equinox API.
            (Keyword only argument.)

        **Returns:**

        The output is a JAX array of the same shape as `x`.

        If `state` is passed, then a 2-tuple of `(output, state)` is returned. The state
        is passed through unchanged. If `state` is not passed, then just the output is
        returned.
        """
        if x.shape != self.shape:
            raise ValueError(
                "`RMSNorm(shape)(x)` must satisfy the invariant `shape == x.shape`"
                f"Received `shape={self.shape} and `x.shape={x.shape}`. You might need "
                "to replace `rms_norm(x)` with `jax.vmap(rms_norm)(x)`.\n"
            )

        orig_dtype = x.dtype

        with jax.numpy_dtype_promotion("standard"):
            dtype = jnp.result_type(x.dtype, jnp.float32)

        x = x.astype(dtype)
        inv_rms = jax.lax.rsqrt(jnp.mean(x**2) + self.eps)
        out = inv_rms * x

        if self.use_weight:
            out = self.weight.astype(dtype) * out  # pyright: ignore
        if self.use_bias:
            out = out + self.bias.astype(dtype)  # pyright: ignore
        if state is sentinel:
            return out.astype(orig_dtype)
        else:
            return out.astype(orig_dtype), state



import warnings
from typing import Any, cast, TYPE_CHECKING, Union

import jax._src.traceback_util as traceback_util
import jax.core
import jax.numpy as jnp
import numpy as np
from jaxtyping import Array, ArrayLike, Bool, Int

from ._doc_utils import doc_repr
from ._errors import branched_error_if
from ._module import field, Module


traceback_util.register_exclusion(__file__)


def _magic(name: str) -> bool:
    return name.startswith("__") and name.endswith("__")


def _store(
    cls,
    name: str,
    message: str,
    name_to_item: dict[str, "EnumerationItem"],
    index_to_message: list[str],
) -> int:
    try:
        our_item = name_to_item[name]
    except KeyError:
        our_index = len(index_to_message)
        name_to_item[name] = EnumerationItem(np.array(our_index, dtype=np.int32), cls)  # pyright: ignore
        index_to_message.append(message)
        return our_index
    else:
        our_index = our_item._value.item()
        existing_message = index_to_message[our_index]
        if message != existing_message:
            warnings.warn(
                f"Enumeration has duplicate incompatible values for {name}. Using the "
                f"latest, which is '{message}'."
            )
        index_to_message[our_index] = message
        return our_index


class _EnumerationMeta(type):
    def __new__(mcs, cls_name, bases, namespace):
        assert "_name_to_item" not in namespace
        assert "_index_to_message" not in namespace
        assert "_base_offsets" not in namespace
        name_to_item = {}
        index_to_message = []
        base_offsets = {}
        new_namespace = dict(
            _name_to_item=name_to_item,
            _index_to_message=index_to_message,
            _base_offsets=base_offsets,
        )
        if cls_name == "Enumeration" and "Enumeration" not in globals().keys():
            new_namespace["promote"] = namespace.pop("promote")
            new_namespace["where"] = namespace.pop("where")
        else:
            if "promote" in namespace:
                raise ValueError(
                    "Cannot have enumeration item with name `promote`, as this "
                    "conflicts with the classmethod of this name"
                )
            if "where" in namespace:
                raise ValueError(
                    "Cannot have enumeration item with name `where`, as this conflicts "
                    "with the classmethod of this name"
                )
        for name, value in namespace.items():
            if _magic(name):
                new_namespace[name] = value
        cls = super().__new__(mcs, cls_name, bases, new_namespace)
        for original_base in bases:
            # skip Enumeration and object
            for base in original_base.__mro__[:-2]:
                if not issubclass(base, Enumeration):
                    raise ValueError(
                        "Cannot subclass enumerations with non-enumerations."
                    )
                if base not in base_offsets.keys():
                    offsets = np.zeros(len(base), dtype=np.int32)
                    for name, base_item in base._name_to_item.items():
                        base_index = base_item._value.item()
                        message = base._index_to_message[base_index]
                        our_index = _store(
                            cls, name, message, name_to_item, index_to_message
                        )
                        offsets[base_index] = our_index
                    base_offsets[base] = np.asarray(offsets)
        for name, message in namespace.items():
            if not _magic(name):
                _store(cls, name, message, name_to_item, index_to_message)
        if "__doc__" not in cls.__dict__ or cls.__doc__ is None:
            doc_pieces = [
                """An
[enumeration](https://docs.kidger.site/equinox/api/enumerations/), with the
following entries:
"""
            ]
            for name, item in name_to_item.items():
                message = index_to_message[item._value.item()]
                if message == "":
                    doc_pieces.append(f"- `{name}`")
                else:
                    message = message.replace("\n", "\n    ")
                    doc_pieces.append(f"- `{name}`: {message}")
            cls.__doc__ = "\n\n".join(doc_pieces)
        return cls

    def __getattr__(cls, item: str):
        try:
            return cls._name_to_item[item]
        except KeyError as e:
            raise AttributeError from e

    def __getitem__(cls, item) -> str:
        if not isinstance(item, EnumerationItem):
            raise ValueError("Index must be an enumeration item.")
        if item._enumeration is not cls:
            raise ValueError("Enumeration item must be from this enumeration.")
        if isinstance(item._value, jax.core.Tracer):
            return "traced"
        elif isinstance(item._value, (np.ndarray, Array)):
            return cls._index_to_message[item._value.item()]
        else:
            # PyTrees have to be generic wrt leaf type.
            return "unknown"

    def __len__(cls):
        return len(cls._name_to_item)

    def __instancecheck__(cls, value):
        return isinstance(value, EnumerationItem) and value._enumeration is cls


class EnumerationItem(Module):
    _value: Int[Union[Array, np.ndarray], ""]
    # Should have annotation `"type[Enumeration]"`, but this fails due to beartype bug
    # #289.
    _enumeration: Any = field(static=True)

    if TYPE_CHECKING:

        def __init__(self, x):
            pass

    def __eq__(self, other) -> Bool[Array, ""]:
        if isinstance(other, EnumerationItem):
            if self._enumeration is other._enumeration:
                with jax.ensure_compile_time_eval():
                    return jnp.asarray(self._value == other._value)
        raise ValueError(
            "Can only compare equality between enumerations of the same type."
        )

    def __ne__(self, other) -> Bool[Array, ""]:  # pyright: ignore
        if isinstance(other, EnumerationItem):
            if self._enumeration is other._enumeration:
                with jax.ensure_compile_time_eval():
                    return jnp.asarray(self._value != other._value)
        raise ValueError(
            "Can only compare equality between enumerations of the same type."
        )

    def __repr__(self):
        prefix = f"{self._enumeration.__module__}.{self._enumeration.__qualname__}"
        message = self._enumeration[self]
        return f"{prefix}<{message}>"

    def error_if(self, token, pred):
        """Conditionally raise a runtime error, with message given by this enumeration
        item.

        See [`equinox.error_if`][] for more information on the behaviour of errors.

        **Arguments:**

        - `token`: the token to thread the error on to. This should be a PyTree
            containing at least one array. The error will be checked after this array
            has been computed, and before the return value from this function is used.
        - `pred`: the condition to raise the error on. Will raise if this evaluates to
            True at runtime.

        **Returns:**

        `token` is returned unchanged.
        """
        return branched_error_if(
            token, pred, self._value, self._enumeration._index_to_message
        )

    def is_traced(self) -> bool:
        return isinstance(self._value, jax.core.Tracer)


if TYPE_CHECKING:
    import enum
    from typing import ClassVar
    from typing_extensions import Self

    class _Sequence(type):
        def __getitem__(cls, item) -> str: ...

        def __len__(cls) -> int: ...

    class Enumeration(enum.Enum, EnumerationItem, metaclass=_Sequence):  # pyright: ignore
        _name_to_item: ClassVar[dict[str, EnumerationItem]]  # pyright: ignore
        _index_to_message: ClassVar[list[str]]  # pyright: ignore
        _base_offsets: ClassVar[dict["Enumeration", int]]

        @classmethod
        def promote(cls, item: "Enumeration") -> Self: ...

        @classmethod
        def where(cls, pred: Bool[ArrayLike, "..."], a: Self, b: Self) -> Self: ...

else:
    _Enumeration = doc_repr(Any, "Enumeration")

    class Enumeration(metaclass=_EnumerationMeta):
        """JAX-compatible enums.

        Enumerations are instantiated using class syntax, and values looked up on the
        class:
        ```python
        class RESULTS(eqx.Enumeration):
            success = "Hurrah!"
            linear_solve_failed = "Linear solve failed to converge"
            diffeq_solve_failed = "Differential equation solve exploded horribly"

        result = RESULTS.success
        ```

        Enumerations support equality checking:
        ```python
        x = jnp.where(result == RESULTS.success, a, b)
        ```

        Enumerations cannot be compared against anything except an Enumeration of the
        same type:
        ```python
        result == 0  # ValueError at trace time: `0` is an integer, not an Enumeration.
        result == SOME_OTHER_ENUMERATION.foo  # Likewise, also a ValueError.
        ```

        Enumerations can be passed through JIT:
        ```python
        jax.jit(lambda x: x)(RESULTS.success)
        ```

        Enumerations use their assigned value in their repr:
        ```python
        print(RESULTS.success)  # RESULTS<Hurrah!>
        ```

        Given a Enumeration element, just the string can be looked up by indexing it:
        ```python
        result = RESULTS.success
        print(RESULTS[result])  # Hurrah!
        ```

        Enumerations support inheritance, to include all of the superclasses' fields as
        well as any new ones. Note that you will need add a `# pyright: ignore` wherever
        you inherit.
        ```python
        class RESULTS(eqx.Enumeration):
            success = "success"
            linear_solve_failed = "Linear solve failed to converge"
            diffeq_solve_failed = "Differential equation solve exploded horribly"

        class MORE_RESULTS(RESULTS):  # pyright: ignore
            flux_capacitor_overloaded = "Run for your life!"

        result = MORE_RESULTS.linear_solve_failed
        ```

        Enumerations are often used to represent error conditions. As such they have
        built-in support for raising runtime errors, via [`equinox.error_if`][]:
        ```python
        x = result.error_if(x, pred)
        ```
        this is equivalent to `x = eqx.error_if(x, pred, msg)`, where `msg` is the
        string corresponding to the enumeration item.
        """

        @classmethod
        def promote(cls, item: _Enumeration) -> _Enumeration:
            """Enums support `.promote` (on the class) to promote from an inherited
            class.

            !!! Example

                ```python
                class RESULTS(eqx.Enumeration):
                    success = "success"
                    linear_solve_failed = "Linear solve failed to converge"
                    diffeq_solve_failed = "Differential equation solve exploded horribly"

                class MORE_RESULTS(RESULTS):  # pyright: ignore
                    flux_capacitor_overloaded = "Run for your life!"

                result == RESULTS.success

                # This is a ValueError at trace time
                result == MORE_RESULTS.success

                # This works. You can only promote from superclasses to subclasses.
                result = MORE_RESULTS.promote(result)
                result == MORE_RESULTS.success
                ```

            **Arguments:**

            - `item`: an item from a parent Enumeration.

            **Returns:**

            `item`, but as a member of this Enumeration.
            """  # noqa: E501
            if not isinstance(item, EnumerationItem):
                raise ValueError("Can only promote enumerations.")
            if (not issubclass(cls, item._enumeration)) or item._enumeration is cls:
                raise ValueError("Can only promote from inherited enumerations.")
            with jax.ensure_compile_time_eval():
                value = jnp.asarray(cls._base_offsets[item._enumeration])[item._value]
            return EnumerationItem(value, cls)

        @classmethod
        def where(
            cls, pred: Bool[ArrayLike, "..."], a: _Enumeration, b: _Enumeration
        ) -> _Enumeration:
            """Enumerations support `.where` (on the class), analogous to `jnp.where`.

            !!! Example

                ```python
                result = RESULTS.where(diff < tol, RESULTS.success, RESULTS.linear_solve_failed)
                result = RESULTS.where(step < max_steps, result, RESULTS.diffeq_solve_failed)
                ```

            **Arguments:**

            - `pred`: a scalar boolean array.
            - `a`: an item of the enumeration. Must be of the same Enumeration as
                `.where` is accessed from.
            - `b`: an item of the enumeration. Must be of the same Enumeration as
                `.where` is accessed from.

            **Returns:**

            `a` if `pred` is true, and `b` is `pred` is false.
            """  # noqa: E501
            if jnp.shape(pred) != () or jnp.result_type(pred) != jnp.bool_:
                raise ValueError("`where` requires a scalar boolean predicate")
            if isinstance(a, EnumerationItem) and isinstance(b, EnumerationItem):
                if a._enumeration is cls and b._enumeration is cls:
                    if a._value is b._value:
                        return a
                    else:
                        with jax.ensure_compile_time_eval():
                            value = jnp.where(pred, a._value, b._value)
                        cls = cast(type[Enumeration], cls)
                        return EnumerationItem(value, cls)
            name = f"{cls.__module__}.{cls.__qualname__}"
            raise ValueError(
                f"Arguments to {name}.where(...) must be members of {name}."
            )



import operator
from collections.abc import Callable
from typing import Any, Optional, TYPE_CHECKING

import jax
import jax.tree_util as jtu


class _Metaω(type):
    def __rpow__(cls, value):
        return cls(value)


class _ω(metaclass=_Metaω):
    """Provides friendlier syntax for mapping with `jax.tree_util.tree_map`.

    !!! example

        ```python
        (ω(a) + ω(b)).ω == jax.tree_util.tree_map(operator.add, a, b)
        ```

    !!! tip

        To minimise the number of brackets, the following `__rpow__` syntax can be
        used:

        ```python
        (a**ω + b**ω).ω == jax.tree_util.tree_map(operator.add, a, b)
        ```

        This is entirely equivalent to the above.
    """

    def __init__(self, value, is_leaf=None):
        """
        **Arguments:**

        - `value`: The PyTree to wrap.
        - `is_leaf`: An optional value for the `is_leaf` argument to
          `jax.tree_util.tree_map`.

        !!! note

            The `is_leaf` argument cannot be set when using the `__rpow__` syntax for
            initialisation.
        """
        self.ω = value
        self.is_leaf = is_leaf

    def __repr__(self):
        return f"ω({self.ω})"

    def __getitem__(self, item):
        return ω(
            jtu.tree_map(lambda x: x[item], self.ω, is_leaf=self.is_leaf),
            is_leaf=self.is_leaf,
        )

    def call(self, fn):
        return ω(
            jtu.tree_map(fn, self.ω, is_leaf=self.is_leaf),
            is_leaf=self.is_leaf,
        )

    @property
    def at(self):
        return _ωUpdateHelper(self.ω, self.is_leaf)


if TYPE_CHECKING:
    ω: Any = ...
else:
    ω = _ω


def _equal_code(fn1: Optional[Callable], fn2: Optional[Callable]):
    """Checks whether fn1 and fn2 both have the same code.

    It's essentially impossible to see if two functions are equivalent, so this won't,
    and isn't intended, to catch every possible difference between fn1 and fn2. But it
    should at least catch the common case that `is_leaf` is specified for one input and
    not specified for the other.
    """
    sentinel1 = object()
    sentinel2 = object()
    code1 = getattr(getattr(fn1, "__code__", sentinel1), "co_code", sentinel2)
    code2 = getattr(getattr(fn2, "__code__", sentinel1), "co_code", sentinel2)
    return type(code1) == type(code2) and code1 == code2


def _set_binary(base, name: str, op: Callable[[Any, Any], Any]) -> None:
    def fn(self, other):
        if isinstance(other, ω):
            if jtu.tree_structure(self.ω) != jtu.tree_structure(other.ω):
                raise ValueError("PyTree structures must match.")
            if not _equal_code(self.is_leaf, other.is_leaf):
                raise ValueError("`is_leaf` must match.")
            return ω(
                jtu.tree_map(op, self.ω, other.ω, is_leaf=self.is_leaf),
                is_leaf=self.is_leaf,
            )
        elif isinstance(other, (bool, complex, float, int, jax.Array)):
            return ω(
                jtu.tree_map(lambda x: op(x, other), self.ω, is_leaf=self.is_leaf),
                is_leaf=self.is_leaf,
            )
        else:
            raise RuntimeError("Type of `other` not understood.")

    fn.__name__ = name
    fn.__qualname__ = base.__qualname__ + "." + name
    setattr(base, name, fn)


def _set_unary(base, name: str, op: Callable[[Any], Any]) -> None:
    def fn(self):
        return ω(
            jtu.tree_map(op, self.ω, is_leaf=self.is_leaf),
            is_leaf=self.is_leaf,
        )

    fn.__name__ = name
    fn.__qualname__ = base.__qualname__ + "." + name
    setattr(base, name, fn)


def _rev(op):
    def __rev(x, y):
        return op(y, x)

    return __rev


for name, op in [
    ("__add__", operator.add),
    ("__sub__", operator.sub),
    ("__mul__", operator.mul),
    ("__matmul__", operator.matmul),
    ("__truediv__", operator.truediv),
    ("__floordiv__", operator.floordiv),
    ("__mod__", operator.mod),
    ("__pow__", operator.pow),
    ("__lshift__", operator.lshift),
    ("__rshift__", operator.rshift),
    ("__and__", operator.and_),
    ("__xor__", operator.xor),
    ("__or__", operator.or_),
    ("__radd__", _rev(operator.add)),
    ("__rsub__", _rev(operator.sub)),
    ("__rmul__", _rev(operator.mul)),
    ("__rmatmul__", _rev(operator.matmul)),
    ("__rtruediv__", _rev(operator.truediv)),
    ("__rfloordiv__", _rev(operator.floordiv)),
    ("__rmod__", _rev(operator.mod)),
    ("__rpow__", _rev(operator.pow)),
    ("__rlshift__", _rev(operator.lshift)),
    ("__rrshift__", _rev(operator.rshift)),
    ("__rand__", _rev(operator.and_)),
    ("__rxor__", _rev(operator.xor)),
    ("__ror__", _rev(operator.or_)),
    ("__lt__", operator.lt),
    ("__le__", operator.le),
    ("__eq__", operator.eq),
    ("__ne__", operator.ne),
    ("__gt__", operator.gt),
    ("__ge__", operator.ge),
]:
    _set_binary(ω, name, op)


for name, op in [
    ("__neg__", operator.neg),
    ("__pos__", operator.pos),
    ("__abs__", operator.abs),
    ("__invert__", operator.invert),
]:
    _set_unary(ω, name, op)


class _ωUpdateHelper:
    def __init__(self, value, is_leaf):
        self.value = value
        self.is_leaf = is_leaf

    def __getitem__(self, item):
        return _ωUpdateRef(self.value, item, self.is_leaf)


class _ωUpdateRef:
    def __init__(self, value, item, is_leaf):
        self.value = value
        self.item = item
        self.is_leaf = is_leaf


def _set_binary_at(base, name: str, op: Callable[[Any, Any, Any], Any]) -> None:
    def fn(self, other):
        if isinstance(other, ω):
            if jtu.tree_structure(self.value) != jtu.tree_structure(other.ω):
                raise ValueError("PyTree structures must match.")
            if not _equal_code(self.is_leaf, other.is_leaf):
                raise ValueError("is_leaf specifications must match.")
            return ω(
                jtu.tree_map(
                    lambda x, y: op(x, self.item, y),
                    self.value,
                    other.ω,
                    is_leaf=self.is_leaf,
                ),
                is_leaf=self.is_leaf,
            )
        elif isinstance(other, (bool, complex, float, int, jax.Array)):
            return ω(
                jtu.tree_map(
                    lambda x: op(x, self.item, other), self.value, is_leaf=self.is_leaf
                ),
                is_leaf=self.is_leaf,
            )
        else:
            raise RuntimeError("Type of `other` not understood.")

    fn.__name__ = name
    fn.__qualname__ = base.__qualname__ + "." + name
    setattr(base, name, fn)


for name, op in [
    ("set", lambda x, y, z, **kwargs: x.at[y].set(z, **kwargs)),
    ("add", lambda x, y, z, **kwargs: x.at[y].add(z, **kwargs)),
    ("multiply", lambda x, y, z, **kwargs: x.at[y].multiply(z, **kwargs)),
    ("divide", lambda x, y, z, **kwargs: x.at[y].divide(z, **kwargs)),
    ("power", lambda x, y, z, **kwargs: x.at[y].power(z, **kwargs)),
    ("min", lambda x, y, z, **kwargs: x.at[y].min(z, **kwargs)),
    ("max", lambda x, y, z, **kwargs: x.at[y].max(z, **kwargs)),
]:
    _set_binary_at(_ωUpdateRef, name, op)



def deprecated_0_10(kwargs, name):
    try:
        kwargs[name]
    except KeyError:
        pass
    else:
        raise ValueError(
            f"{name} is deprecated as of Equinox version 0.10.0. Please see the "
            "release notes on GitHub for details on how to upgrade:\n"
            "https://github.com/patrick-kidger/equinox/releases/tag/v0.10.0"
        )



import jax
import jax._src.traceback_util as traceback_util
import jax.core
import jax.lax as lax
from jaxtyping import Array, Bool

from .._unvmap import unvmap_any


traceback_util.register_exclusion(__file__)


def breakpoint_if(pred: Bool[Array, "..."], **kwargs):
    """As `jax.debug.breakpoint`, but only triggers if `pred` is True.

    **Arguments:**

    - `pred`: the predicate for whether to trigger the breakpoint.
    - `**kwargs`: any other keyword arguments to forward to `jax.debug.breakpoint`.

    **Returns:**

    Nothing.
    """
    # We can't just write `jax.debug.breakpoint` for the second branch. For some reason
    # it needs as lambda wrapper.
    token = kwargs.get("token", None)
    return lax.cond(
        unvmap_any(pred), lambda: jax.debug.breakpoint(**kwargs), lambda: token
    )



import dataclasses
import functools as ft
import inspect
import warnings
from collections.abc import Callable, Hashable
from typing import Any, Literal, Optional, overload, Union

import jax
import jax._src.traceback_util as traceback_util
import jax.core
import jax.interpreters.batching as batching
import jax.interpreters.pxla as pxla
import jax.numpy as jnp
import jax.tree_util as jtu
import numpy as np
from jaxtyping import PyTree

from ._compile_utils import (
    compile_cache,
    hashable_combine,
    hashable_partition,
    Lowered,
)
from ._custom_types import sentinel
from ._deprecate import deprecated_0_10
from ._doc_utils import doc_remove_args
from ._filters import combine, filter, is_array, is_array_like, partition
from ._module import Module, module_update_wrapper, Partial, Static


traceback_util.register_exclusion(__file__)


ResolvedAxisSpec = Optional[int]
AxisSpec = Union[ResolvedAxisSpec, Callable[[Any], ResolvedAxisSpec]]


def _is_none(x: Any) -> bool:
    return x is None


def _resolve_axis(axis_spec: AxisSpec, elem: Any) -> PyTree[ResolvedAxisSpec]:
    if axis_spec is None or isinstance(axis_spec, int):
        return jtu.tree_map(lambda _: axis_spec, elem)
    elif callable(axis_spec):
        return jtu.tree_map(axis_spec, elem)
    else:
        raise ValueError(
            "`in_axes` and `out_axes` must consist of None, ints, and callables only."
        )


def _resolve_axes(
    pytree: PyTree[Any], axes_spec: PyTree[AxisSpec]
) -> PyTree[ResolvedAxisSpec]:
    return jtu.tree_map(_resolve_axis, axes_spec, pytree, is_leaf=_is_none)


@dataclasses.dataclass(frozen=True)  # not a pytree
class if_array:
    """Returns a callable that returns the specified integer if evaluated on an array.
    Otherwise, it returns `None`.

    !!! Example

        ```python
        fn = if_array(1)
        # Evaluate on an array, return the integer.
        fn(jax.numpy.array([0, 1, 2]))  # 1
        # Evaluate on not-an-array, return None.
        fn(True)  # None
        ```
    """

    axis: int

    def __call__(self, x: Any) -> Optional[int]:
        return self.axis if is_array(x) else None


@dataclasses.dataclass(frozen=True)  # not a pytree
class if_mapped:
    """Used with the `out_axes` argument of [`equinox.filter_vmap`][], to only add an
    output batch axis if necessary.

    That is, `out_axes=if_mapped(i)` is equivalent to `out_axes=i` for any output that
    is batched, and `out_axes=None` fofr any output that is not batched.
    """

    axis: int

    def __call__(self, x: Any):
        raise RuntimeError(
            "`eqx.internal.if_mapped` should not be called directly; it is only valid "
            "when passed to `out_axes` of `eqx.filter_vmap`."
        )


@dataclasses.dataclass(frozen=True)  # not a pytree
class _if_mapped:
    main: Any
    axis: int

    def __call__(self, x: Any) -> Optional[int]:
        if isinstance(x, batching.BatchTracer) and x._trace.main is self.main:
            if x.batch_dim is batching.not_mapped:
                return None
            else:
                return self.axis
        elif isinstance(x, pxla.MapTracer) and x._trace.main is self.main:
            return self.axis
        else:
            return None


# The existence of this function is a complete hack: it couples together `filter_vmap`
# with `if_mapped`. I don't see an obvious way around it though.
def _bind_main(main, out_axes):
    def _bind(axis):
        if isinstance(axis, if_mapped):
            return _if_mapped(main, axis.axis)
        else:
            return axis

    return jtu.tree_map(_bind, out_axes)


def _swapaxes(array, axis):
    return jnp.swapaxes(array, 0, axis)


def _named_in_axes(fun, in_axes, args):
    if isinstance(in_axes, dict):
        in_axes = dict(in_axes)
        new_in_axes = []
        default = if_array(0)
        params = inspect.signature(fun).parameters
        # We may have that len(args) < len(params) due to default arguments.
        # Truncate to considering just the arguments that have been passed.
        #
        # (If len(args) > len(params) then we'll get the usual error later when
        # attempting to call with the wrong number of arguments.)
        for _, param_name in zip(args, params):
            new_in_axes.append(in_axes.pop(param_name, default))
        if len(in_axes) != 0:
            raise ValueError(
                "The following `in_axes` did not correspond to any argument: "
                f"{tuple(in_axes.keys())}"
            )
        # Note that this requires all named arguments to be passed; they cannot take
        # default values. That is, we deliberately don't allow something like
        # ```python
        # @eqx.filter_vmap(in_axes=dict(foo=0))
        # def fn(foo=default_value):
        #     ...
        #
        # fn()
        # ```
        # This is because it is ambiguous whether the default value is vectorised or
        # not.
        in_axes = tuple(new_in_axes)
    return in_axes


class _VmapWrapper(Module):
    _fun: Callable
    _in_axes: PyTree[AxisSpec]
    _out_axes: PyTree[AxisSpec]
    _axis_name: Optional[Hashable]
    _axis_size: Optional[int]
    _vmapkwargs: dict[str, Any]

    @property
    def __wrapped__(self):
        return self._fun

    def __call__(self, /, *args, **kwargs):
        if len(kwargs) != 0:
            raise RuntimeError(
                "keyword arguments cannot be used with functions wrapped with "
                "`filter_vmap`"
            )
        del kwargs

        in_axes = _named_in_axes(self._fun, self._in_axes, args)

        # JAX only actually supports passing array-typed values as inputs.
        # Interestingly it's usually still fine to pass non-array-typed values as
        # broadcast inputs. *Unless* you mess up with an inconsistent axis size, in
        # which case it tries to get the axis sizes of all inputs and then fails on the
        # non-array inputs.
        # e.g.:
        # ```
        # jax.vmap(lambda x, y, z: 0,
        #          in_axes=(0, 0, None))(jnp.arange(3), jnp.arange(5), object())
        # ```
        in_axes = _resolve_axes(args, in_axes)
        unmapped_axis = jtu.tree_map(_is_none, in_axes, is_leaf=_is_none)
        static_args, dynamic_args = partition(args, unmapped_axis)

        def _fun_wrapper(_dynamic_args):
            _main = jax.core.find_top_trace(jtu.tree_leaves(_dynamic_args)).main
            _args = combine(_dynamic_args, static_args)
            _out = self._fun(*_args)
            _out_axes = _bind_main(_main, self._out_axes)
            _out_axes = _resolve_axes(_out, _out_axes)
            _none_axes = jtu.tree_map(_is_none, _out_axes, is_leaf=_is_none)
            _nonvmapd, _vmapd = partition(_out, _none_axes, is_leaf=_is_none)
            _nonvmapd_arr, _nonvmapd_static = partition(_nonvmapd, is_array)
            return _vmapd, _nonvmapd_arr, Static((_nonvmapd_static, _out_axes))

        if len(jtu.tree_leaves(in_axes)) == 0 and self._axis_size is None:
            vmapd, nonvmapd_arr, static = _fun_wrapper(dynamic_args)
            if len(jtu.tree_leaves(vmapd)) != 0:
                raise ValueError(
                    "Cannot resolve batch dimension. Non-`None` `out_axes` requires "
                    "either `in_axes` or `axis_size` to be not `None`."
                )
        else:
            vmapd, nonvmapd_arr, static = jax.vmap(
                _fun_wrapper,
                in_axes=(in_axes,),
                out_axes=(0, None, None),
                axis_name=self._axis_name,
                axis_size=self._axis_size,
                **self._vmapkwargs,
            )(dynamic_args)

        nonvmapd_static, out_axes = static.value
        nonvmapd = combine(nonvmapd_arr, nonvmapd_static)

        assert jtu.tree_structure(vmapd) == jtu.tree_structure(out_axes)
        vmapd = jtu.tree_map(_swapaxes, vmapd, out_axes)

        return combine(vmapd, nonvmapd)

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return Partial(self, instance)


@overload
def filter_vmap(
    *,
    in_axes: PyTree[AxisSpec] = if_array(0),
    out_axes: PyTree[AxisSpec] = if_array(0),
    axis_name: Hashable = None,
    axis_size: Optional[int] = None,
) -> Callable[[Callable[..., Any]], Callable[..., Any]]: ...


@overload
def filter_vmap(
    fun: Callable[..., Any],
    *,
    in_axes: PyTree[AxisSpec] = if_array(0),
    out_axes: PyTree[AxisSpec] = if_array(0),
    axis_name: Hashable = None,
    axis_size: Optional[int] = None,
) -> Callable[..., Any]: ...


@doc_remove_args("vmapkwargs")
def filter_vmap(
    fun=sentinel,
    *,
    in_axes: PyTree[AxisSpec] = if_array(0),
    out_axes: PyTree[AxisSpec] = if_array(0),
    axis_name: Hashable = None,
    axis_size: Optional[int] = None,
    **vmapkwargs,
):
    """Vectorises a function. By default, all JAX/NumPy arrays are vectorised down their
    leading axis (i.e. axis index 0), and all other types are broadcast.

    **Arguments:**

    For both `in_axes` and `out_axes`, then `int` indicates an array axis to vectorise
    over, `None` indicates that an argument should be broadcast (not vectorised
    over), and callables `Leaf -> Union[None, int]` are mapped and evaluated on every
    leaf of their subtree. `None` should be used for non-JAX-array arguments.

    - `fun` is a pure function to vectorise. Should be of the form `fun(*args)`; that
        is to say it cannot accept keyword arguments.
    - `in_axes` indicates which axes of the input arrays should be vectorised over.
        It should be a PyTree of `None`, `int`, or callables `Leaf -> Union[None, int]`.
        Its tree structure should either be:
        1. a prefix of the input tuple of `args`.
        2. a dictionary, in which case the named arguments use the specified indices
            to vectorise over, and all other arguments will have the default
            `eqx.if_array(0)`.
    - `out_axes` indicates which axis of the output arrays the mapped axis should appear
        at. It should be a PyTree of `None`, `int`, or callables
        `Leaf -> Union[None, int]`, and its tree structure should be a prefix of the
        output `fun(*args)`.
    - `axis_name` is an optional hashable Python object used to identify the mapped
        axis so that parallel collectives (e.g. `jax.lax.psum`) can be applied.
    - `axis_size` is an optional `int` describing the size of the axis mapped. This
        only needs to be passed if none of the input arguments are vectorised, as else
        it can be deduced by looking at the argument shapes.

    **Returns:**

    The vectorised version of `fun`.

    !!! tip

        To vectorise all JAX/NumPy arrays down their `j`th axis, and broadcast all other
        types, then you can use `equinox.if_array(j)`, which returns a callable
        `leaf -> j if is_array(leaf) else None`. For example: the default values of
        `in_axes` and `out_axes` are both `equinox.if_array(0)`.

    !!! example

        ```python
        import equinox as eqx
        import jax.numpy as jnp

        @eqx.filter_vmap
        def f(x, y):
            return x + y

        @eqx.filter_vmap(in_axes=(None, 1))
        def g(x, y):
            return x + y

        f(jnp.array([1, 2]), jnp.array([3, 4]))  # both args vectorised down axis 0

        f(jnp.array([1, 2]), 3)                  # first arg vectorised down axis 0
                                                 # second arg broadcasted

        g(jnp.array(1), jnp.array([[2, 3]]))     # first arg broadcasted
                                                 # second arg vectorised down axis 1
        ```

    !!! example

        `filter_vmap` can be used to easily create ensembles of models. For example,
        here's an ensemble of eight MLPs:

        ```python
        import equinox as eqx
        import jax.random as jr

        key = jr.PRNGKey(0)
        keys = jr.split(key, 8)

        # Create an ensemble of models

        @eqx.filter_vmap
        def make_ensemble(key):
            return eqx.nn.MLP(2, 2, 2, 2, key=key)

        mlp_ensemble = make_ensemble(keys)

        # Evaluate each member of the ensemble on the same data

        @eqx.filter_vmap(in_axes=(eqx.if_array(0), None))
        def evaluate_ensemble(model, x):
            return model(x)

        evaluate_ensemble(mlp_ensemble, jr.normal(key, (2,)))

        # Evaluate each member of the ensemble on different data

        @eqx.filter_vmap
        def evaluate_per_ensemble(model, x):
            return model(x)

        evaluate_per_ensemble(mlp_ensemble, jr.normal(key, (8, 2)))
        ```

        Here, `make_ensemble` works because [`equinox.nn.MLP`][] is a PyTree, and so it
        is a valid output from a `filter_vmap`. This PyTree includes some JAX arrays
        (the weights and biases) and some non-JAX-arrays (e.g. activation functions).
        `filter_vmap` will vectorise the JAX arrays (with separate weights for each
        member of the ensemble) whilst leaving the non-JAX-arrays alone.

        Note that as the weights in `mlp_ensemble` now have a leading batch dimension
        -- that the weights of `eqx.nn.MLP` instances do not typically have -- then it
        cannot be called directly. It must instead be passed back into a vectorised
        region to be called.
    """

    if fun is sentinel:
        return ft.partial(
            filter_vmap,
            in_axes=in_axes,
            out_axes=out_axes,
            axis_name=axis_name,
            axis_size=axis_size,
            **vmapkwargs,
        )

    deprecated_0_10(vmapkwargs, "default")
    deprecated_0_10(vmapkwargs, "fn")
    deprecated_0_10(vmapkwargs, "args")
    deprecated_0_10(vmapkwargs, "kwargs")
    deprecated_0_10(vmapkwargs, "out")

    vmap_wrapper = _VmapWrapper(
        _fun=fun,
        _in_axes=in_axes,
        _out_axes=out_axes,
        _axis_name=axis_name,
        _axis_size=axis_size,
        _vmapkwargs=vmapkwargs,
    )
    return module_update_wrapper(vmap_wrapper)


@compile_cache
def _filter_pmap_cache(
    fun_names, static, struct, in_axes, axis_name, axis_size, pmapkwargs
):
    def fun_abstract(_dynamic):
        fun, args, _, _ = combine(_dynamic, static)
        out = fun(*args)
        return filter(out, is_array_like)

    fun_abstract = jax.vmap(
        fun_abstract, in_axes=(in_axes,), axis_name=axis_name, axis_size=axis_size
    )
    struct_out = jax.eval_shape(fun_abstract, struct)
    max_out_size = jtu.tree_reduce(lambda x, y: max(x, y.ndim), struct_out, 0)
    del fun_abstract, struct, struct_out

    def _check_map_out_axis(x: Optional[int]):
        if isinstance(x, int):
            if x < -max_out_size or x >= max_out_size:
                raise ValueError(
                    "integers in filter_pmap(..., out_axes=...) must correspond to a "
                    "dimension of the output array"
                )
        elif x is not None:
            raise ValueError(
                "filter_pmap(..., out_axes=...) must contain only integers and Nones"
            )

    def fun_wrapped(_dynamic):
        _main = jax.core.find_top_trace(jtu.tree_leaves(_dynamic))
        _fun, _args, _, _out_axes = combine(_dynamic, static)
        _out = _fun(*_args)
        _out_axes = _bind_main(_main, _out_axes)
        _out_axes = _resolve_axes(_out, _out_axes)
        jtu.tree_map(_check_map_out_axis, _out_axes)
        _pmapd = []
        for i in range(-max_out_size, max_out_size):
            _i_axes = jtu.tree_map(lambda a: a == i, _out_axes, is_leaf=_is_none)
            _pmapd.append(filter(_out, _i_axes, is_leaf=_is_none))
        _none_axes = jtu.tree_map(_is_none, _out_axes, is_leaf=_is_none)
        _nonpmapd = filter(_out, _none_axes, is_leaf=_is_none)
        _dynamic_nonpmapd, _static_nonpmapd = hashable_partition(_nonpmapd, is_array)
        return _pmapd, _dynamic_nonpmapd, Static(_static_nonpmapd)

    fun_name, fun_qualname = fun_names
    fun_wrapped.__name__ = fun_name
    fun_wrapped.__qualname__ = fun_qualname

    out_axes = (list(range(-max_out_size, max_out_size)), None, None)

    return jax.pmap(
        fun_wrapped,
        in_axes=(in_axes,),  # pyright: ignore
        out_axes=out_axes,  # pyright: ignore
        axis_name=axis_name,
        axis_size=axis_size,
        **pmapkwargs,
    )


def _common_preprocess(axis_size, kwargs):
    if len(kwargs) != 0:
        raise RuntimeError(
            "keyword arguments cannot be used with functions wrapped with "
            "`filter_pmap`"
        )
    if axis_size is None:
        return 0  # hashable non-array object
    else:
        # Work around JAX issue #9252
        return np.broadcast_to(0, axis_size)


def _preprocess(info, args, kwargs):
    fun, out_axes, axis_size = info
    maybe_dummy = _common_preprocess(axis_size, kwargs)
    del kwargs
    dynamic = filter((fun, args, maybe_dummy, out_axes), is_array)
    return (dynamic,)


def _postprocess(out):
    (pmapd, dynamic_nonpmapd, static_nonpmapd) = out
    nonpmapd = hashable_combine(dynamic_nonpmapd, static_nonpmapd.value)
    return combine(*pmapd, nonpmapd)


class _PmapWrapper(Module):
    _fun: Callable
    _in_axes: PyTree[AxisSpec]
    _out_axes: PyTree[AxisSpec]
    _axis_name: Optional[Hashable]
    _axis_size: Optional[int]
    _filter_warning: bool
    _pmapkwargs: dict[str, Any]

    @property
    def __wrapped__(self):
        return self._fun

    def _call(self, is_lower, args, kwargs):
        maybe_dummy = _common_preprocess(self._axis_size, kwargs)
        del kwargs

        in_axes = _named_in_axes(self._fun, self._in_axes, args)
        in_axes = _resolve_axes(args, in_axes)
        in_axes = (None, in_axes, 0, None)

        dynamic, static = partition(
            (self._fun, args, maybe_dummy, self._out_axes), is_array
        )
        struct = jtu.tree_map(lambda x: jax.ShapeDtypeStruct(x.shape, x.dtype), dynamic)

        cached = _filter_pmap_cache(
            self._fun,
            static,
            struct,
            in_axes,
            self._axis_name,
            self._axis_size,
            self._pmapkwargs,
        )

        if is_lower:
            return Lowered(
                cached.lower(dynamic),
                (self._fun, self._out_axes, self._axis_size),
                _preprocess,  # pyright: ignore
                _postprocess,  # pyright: ignore
            )
        else:
            if self._filter_warning is True:
                with warnings.catch_warnings():
                    warnings.filterwarnings(
                        "ignore", message="Some donated buffers were not usable*"
                    )
                    out = cached(dynamic)
            else:
                out = cached(dynamic)
            return _postprocess(out)

    def __call__(self, /, *args, **kwargs):
        return self._call(False, args, kwargs)

    def lower(self, /, *args, **kwargs) -> Lowered:
        return self._call(True, args, kwargs)

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return Partial(self, instance)


# Deliberately using `Callable[..., Any]` as `filter_pmap` does change the input and
# out args in ways not expressible in the static type system (changing the number of
# axes).
@overload
def filter_pmap(
    *,
    in_axes: PyTree[AxisSpec] = if_array(0),
    out_axes: PyTree[AxisSpec] = if_array(0),
    axis_name: Hashable = None,
    axis_size: Optional[int] = None,
    donate: Literal["all", "warn", "none"] = "none",
) -> Callable[[Callable[..., Any]], Callable[..., Any]]: ...


@overload
def filter_pmap(
    fun: Callable[..., Any],
    *,
    in_axes: PyTree[AxisSpec] = if_array(0),
    out_axes: PyTree[AxisSpec] = if_array(0),
    axis_name: Hashable = None,
    axis_size: Optional[int] = None,
    donate: Literal["all", "warn", "none"] = "none",
) -> Callable[..., Any]: ...


@doc_remove_args("pmapkwargs")
def filter_pmap(
    fun=sentinel,
    *,
    in_axes: PyTree[AxisSpec] = if_array(0),
    out_axes: PyTree[AxisSpec] = if_array(0),
    axis_name: Hashable = None,
    axis_size: Optional[int] = None,
    donate: Literal["all", "warn", "none"] = "none",
    **pmapkwargs,
):
    """
    !!! warning

        JAX has now added more powerful parallelism APIs directly to the JIT interface.
        As such, using [`equinox.filter_jit`][] with sharded inputs is now recommended
        over `filter_pmap`. See also the
        [parallelism example](../../examples/parallelism/).

    Parallelises a function. By default, all JAX/NumPy arrays are parallelised down
    their leading axis (i.e. axis index 0), and all other types are broadcast.

    `jax.pmap`, and thus `equinox.filter_pmap`, also compiles their function in the same
    way as `jax.jit`. By default, all JAX arrays are traced, and all other arguments are
    treated as static inputs.

    **Arguments:**

    For both `in_axes` and `out_axes`, then `int` indicates an array axis to parallelise
    over, `None` indicates that an argument should be broadcast (not parallelise
    over), and callables `Leaf -> Union[None, int]` are mapped and evaluated on every
    leaf of their subtree. `None` should be used for non-JAX-array arguments.

    - `fun` is a pure function to parallelise. Should be of the form `fun(*args)`; that
        is to say it cannot accept keyword arguments.
    - `in_axes` indicates which axes of the input arrays should be parallelised over.
        It should be a PyTree of `None`, `int`, or callables `Leaf -> Union[None, int]`.
        Its tree structure should either be:
        1. a prefix of the input tuple of `args`.
        2. a dictionary, in which case the named arguments use the specified indices
            to parallelise over, and all other arguments will have the default
            `eqx.if_array(0)`.
    - `out_axes` indicates which axis of the output arrays the mapped axis should appear
        at. It should be a PyTree of `None`, `int`, or callables
        `Leaf -> Union[None, int]`, and its tree structure should be a prefix of the
        output `fun(*args)`.
    - `axis_name` is an optional hashable Python object used to identify the mapped
        axis so that parallel collectives (e.g. `jax.lax.psum`) can be applied.
    - `axis_size` is an optional `int` describing the size of the axis mapped. This
        only needs to be passed if none of the input arguments are vectorised, as else
        it can be deduced by looking at the argument shapes.
    - `donate` indicates whether the buffers of JAX arrays are donated or not, it
        should either be:
        - `'all'`: donate all arrays and suppress all warnings about
            unused buffers;
        - `'warn'`: as above, but don't suppress unused buffer warnings;
        - `'none'`: the default, disables buffer donation.

    **Returns:**

    The parallelised version of `fun`.

    !!! tip

        To parallelise all JAX/NumPy arrays down their `j`th axis, and broadcast all
        other types, then you can use `equinox.if_array(j)`, which returns a callable
        `leaf -> j if is_array(leaf) else None`. For example: the default values of
        `in_axes` and `out_axes` are both `equinox.if_array(0)`.

    !!! example

        ```python
        import equinox as eqx
        import jax.numpy as jnp

        @eqx.filter_pmap
        def f(x, y):
            return x + y

        @eqx.filter_pmap(in_axes=(None, 1))
        def g(x, y):
            return x + y

        f(jnp.array([1, 2]), jnp.array([3, 4]))  # both args parallelised down axis 0

        f(jnp.array([1, 2]), 3)                  # first arg parallelised down axis 0
                                                 # second arg broadcasted (as it's not
                                                 # a JAX array)

        g(jnp.array(1), jnp.array([[2, 3]]))     # first arg broadcasted
                                                 # second arg parallelised down axis 1
        ```
    """

    if fun is sentinel:
        return ft.partial(
            filter_pmap,
            in_axes=in_axes,
            out_axes=out_axes,
            axis_name=axis_name,
            axis_size=axis_size,
            donate=donate,
            **pmapkwargs,
        )

    deprecated_0_10(pmapkwargs, "default")
    deprecated_0_10(pmapkwargs, "fn")
    deprecated_0_10(pmapkwargs, "args")
    deprecated_0_10(pmapkwargs, "kwargs")
    deprecated_0_10(pmapkwargs, "out")
    if any(x in pmapkwargs for x in ("static_broadcasted_argnums", "donate_argnums")):
        raise ValueError(
            "`pmapkwargs` cannot contain either 'static_broadcasted_argnums' or "
            "'donate_argnums'"
        )

    if donate == "arrays":
        warnings.warn(
            "The `donate='arrays'` option to `filter_pmap` has been renamed to "
            "`donate='all'`",
            DeprecationWarning,
        )
        donate = "all"
    if donate not in {"all", "warn", "none"}:
        raise ValueError(
            "`filter_jit(..., donate=...)` must be one of 'all', 'warn', or 'none'"
        )
    filter_warning = True if donate == "all" else False
    if donate != "none":
        pmapkwargs["donate_argnums"] = (0,)

    pmap_wrapper = _PmapWrapper(
        _fun=fun,
        _in_axes=in_axes,
        _out_axes=out_axes,
        _axis_name=axis_name,
        _axis_size=axis_size,
        _filter_warning=filter_warning,
        _pmapkwargs=pmapkwargs,
    )
    return module_update_wrapper(pmap_wrapper)



"""Implements the core Module abstraction.

This includes both the core dataclass+pytree combo, plus various related pieces of
functionality.
"""

import abc
import dataclasses
import functools as ft
import inspect
import types
import warnings
import weakref
from collections.abc import Callable
from typing import Any, cast, Optional, Protocol, TYPE_CHECKING, TypeVar, Union
from typing_extensions import dataclass_transform, ParamSpec

import jax
import jax._src.traceback_util as traceback_util
import jax.tree_util as jtu
import numpy as np
from jaxtyping import Array, Bool, PyTreeDef

from ._better_abstract import ABCMeta, dataclass
from ._caches import cache_clears
from ._doc_utils import doc_repr
from ._filters import is_array, is_array_like
from ._pretty_print import tree_pformat
from ._tree import tree_equal


traceback_util.register_exclusion(__file__)


_P = ParamSpec("_P")
_T = TypeVar("_T")


#
# Part 1: fields. We extend `dataclasses.field` with a wrapper supporting extra features
# like type converters, and marking a field as static metadata.
#


def static_field(**kwargs):
    """Deprecated in favour of [`equinox.field`][], i.e. `eqx.field(static=True)`."""
    return field(**kwargs, static=True)


_converter_sentinel: Any = doc_repr(object(), "lambda x: x")


def field(
    *,
    converter: Callable[[Any], Any] = _converter_sentinel,
    static: bool = False,
    **kwargs,
):
    """Equinox supports extra functionality on top of the default dataclasses.

    **Arguments:**

    - `converter`: a function to call on this field when the model is initialised. For
        example, `field(converter=jax.numpy.asarray)` to convert
        `bool`/`int`/`float`/`complex` values to JAX arrays. This is ran after the
        `__init__` method (i.e. when using a user-provided `__init__`), and before
        `__post_init__` (i.e. when using the default dataclass initialisation).
    - `static`: whether the field should not interact with any JAX transform at all (by
        making it part of the PyTree structure rather than a leaf).
    - `**kwargs`: All other keyword arguments are passed on to `dataclass.field`.

    !!! example "Example for `converter`"

        ```python
        class MyModule(eqx.Module):
            foo: Array = eqx.field(converter=jax.numpy.asarray)

        mymodule = MyModule(1.0)
        assert isinstance(mymodule.foo, jax.Array)
        ```

    !!! example "Example for `static`"

        ```python
        class MyModule(eqx.Module):
            normal_field: int
            static_field: int = eqx.field(static=True)

        mymodule = MyModule("normal", "static")
        leaves, treedef = jax.tree_util.tree_flatten(mymodule)
        assert leaves == ["normal"]
        assert "static" in str(treedef)
        ```

    `static=True` means that this field is not a node of the PyTree, so it does not
    interact with any JAX transforms, like JIT or grad. This means that it is usually a
    bug to make JAX arrays be static fields. `static=True` should very rarely be used.
    It is preferred to just filter out each field with `eqx.partition` whenever you need
    to select only some fields.
    """
    try:
        metadata = dict(kwargs.pop("metadata"))  # safety copy
    except KeyError:
        metadata = {}
    if "converter" in metadata:
        raise ValueError("Cannot use metadata with `static` already set.")
    if "static" in metadata:
        raise ValueError("Cannot use metadata with `static` already set.")
    # We don't just use `lambda x: x` as the default, so that this works:
    # ```
    # class Abstract(eqx.Module):
    #     x: int = eqx.field()
    #
    # class Concrete(Abstract):
    #    @property
    #    def x(self):
    #        pass
    # ```
    # otherwise we try to call the default converter on a property without a setter,
    # and an error is raised.
    # Oddities like the above are to be discouraged, of course, but in particular
    # `field(init=False)` was sometimes used to denote an abstract field (prior to the
    # introduction of `AbstractVar`), so we do want to support this.
    if converter is not _converter_sentinel:
        metadata["converter"] = converter
    if static:
        metadata["static"] = True
    return dataclasses.field(metadata=metadata, **kwargs)


#
# Part 2: Modules!
# This is the core of Equinox.
#


@dataclass(frozen=True)
class StrictConfig:
    """Used to configure strict Equinox modules.

    Passing this as the `strict` parameter of a `Module` marks that the `Module` is
    strict, and additionally provides some extra custom behaviour.

    Example usage is:
    ```python
    class Foo(eqx.Module, strict=StrictConfig(force_abstract=True)):
        pass
    ```
    """

    force_abstract: bool = False
    allow_abstract_name: bool = False
    allow_method_override: bool = False


StrictConfig.__init__.__doc__ = """**Arguments:**

- `force_abstract`: marks that a class without abstract methods or attributes should
    still be treated as abstract. Useful for e.g.
    ```python
    class AbstractFoo(eqx.Module, strict=True):
        pass
    ```
    which would otherwise be treated as concrete.
- `allow_abstract_name`: grants this class an exemption from the rule that abstract
    classes must begin their names with 'Abstract' or '_Abstract', and that concrete
    classes must not begin their names with 'Abstract' or '_Abstract'.
- `allow_method_override`: grants this class an exemption from the rule that it cannot
    override concrete methods.

Both `allow_*` options should be used with care! They exist only to make it easier to
transition a codebase from non-strict to strict `Module`s in a backward-compatible
manner. If you are starting a new codebase you should not have need of them.
"""


# Inherits from ABCMeta to support `eqx.{AbstractVar, AbstractClassVar}` and
# `abc.abstractmethod`.
class _ActualModuleMeta(ABCMeta):
    # This method is called whenever you definite a module: `class Foo(eqx.Module): ...`
    def __new__(
        mcs,
        name,
        bases,
        dict_,
        /,
        strict: Union[bool, StrictConfig] = False,
        **kwargs,
    ):
        if isinstance(strict, bool):
            strict_config = StrictConfig()
        elif isinstance(strict, StrictConfig):
            strict_config = strict
            strict = True
        else:
            raise TypeError(
                "The `strict` parameter in `class Foo(equinox.Module, strict=...)` "
                "must either be a `bool` or `equinox.StrictConfig` object."
            )

        # [Step 1] Create the class as normal.
        cls = super().__new__(mcs, name, bases, dict_, **kwargs)
        # [Step 2] Arrange for bound methods to be treated as PyTrees as well. This
        # ensures that
        # ```
        # @jax.jit
        # def run(fn):
        #     ...
        # run(SomeModule().some_method)
        # ```
        # works.
        for k, v in cls.__dict__.items():
            if _not_magic(k) and inspect.isfunction(v):
                setattr(cls, k, _wrap_method(v))

        # [Step 3] Handle initialisation.
        #
        # For context, with any Python dataclass, there are three possible scenarios for
        # initialisation:
        # (a) a user-provided `__init__` method is supplied;
        # (b) dataclasses creates `__init__` , without a user-provided `__post_init__`
        # (c) dataclasses creates `__init__` , with a user-provided `__post_init__`

        # Create a dataclass `__init__` method if a user method isn't provided.
        # If a user passed one on this class, then we definitely have a custom __init__.
        # Else just use whatever our superclass does. Note that this is different to
        # default dataclass behaviour. Given
        # ```
        # @dataclass
        # class Foo: def __init__(...): ...
        # @dataclass
        # class Bar(Foo): pass
        # ```
        # then `Bar` will end up with a dataclass-provided `__init__`. That ends up
        # being ergonomically very annoying, so we disable it.
        added_custom_init = "__init__" in cls.__dict__
        if added_custom_init:
            has_dataclass_init = False
        else:
            for kls in cls.__mro__[1:-1]:
                try:
                    has_dataclass_init = _has_dataclass_init[kls]
                except KeyError:
                    # Non-Module superclasses.
                    if kls.__init__ is not object.__init__:
                        has_dataclass_init = False
                        break
                else:
                    break
            else:
                assert name == "Module"
                has_dataclass_init = True  # eqx.Module itself

        # Check for a common error. (Check for `_Initable` to avoid duplicate warnings.)
        if (
            not has_dataclass_init
            and hasattr(cls, "__post_init__")
            and not issubclass(cls, _Initable)
        ):
            warnings.warn(
                f"Class `{cls.__module__}.{cls.__qualname__}` has both an "
                "`__init__` method and a `__post_init__` method. This means that "
                "the `__post_init__` method will not be run!\n"
                "The reason for this is that `__post_init__` is intended to be "
                "used with the automatically-generated `__init__` method provided "
                "by Python dataclasses, which are generated of the form:\n"
                "```\n"
                "def __init__(self, field1, field2)\n"
                "    self.field1 = field1\n"
                "    self.field2 = field2\n"
                "    self.__post_init__()\n"
                "```\n"
                "and as such a user-provided `__init__` overrides both the setting "
                "of fields, and the calling of `__post_init__`.\n"
                "The above is how Python dataclasses work, and has nothing "
                "to do with Equinox!\n"
                "If you are using `__post_init__` to check that certain invariants "
                "hold, then consider using `__check_init__` instead. This is an "
                "Equinox-specific extension that is always ran. See here for more "
                "details: "
                "https://docs.kidger.site/equinox/api/module/advanced_fields/#checking-invariants",  # noqa: E501
                stacklevel=2,
            )

        # Add support for `eqx.field(converter=...)` when using `__post_init__`.
        # (Scenario (c) above. Scenarios (a) and (b) are handled later.)
        if has_dataclass_init and "__post_init__" in cls.__dict__:
            post_init = cls.__post_init__

            @ft.wraps(post_init)  # pyright: ignore
            def __post_init__(self, *args, **kwargs):
                # This `if` is to handle `super()` correctly.
                # We want to only convert once, at the top level.
                #
                # This check is basically testing whether or not the function we're in
                # now (`cls`) is at the top level (`self.__class__`). If we are, do
                # conversion. If we're not, it's presumably because someone is calling
                # us via `super()` in the middle of their own `__post_init__`. No
                # conversion then; their own version of this wrapper will do it at the
                # appropriate time instead.
                #
                # This top-level business means that this is very nearly the same as
                # doing conversion in `_ModuleMeta.__call__`. The differences are that
                # (a) that wouldn't allow us to convert fields before the user-provided
                # `__post_init__`, and (b) it allows other libraries (i.e. jaxtyping)
                # to later monkey-patch `__init__`, and we have our converter run before
                # their own monkey-patched-in code.
                if self.__class__ is _make_initable_wrapper(cls):
                    # Convert all fields currently available.
                    _convert_fields(self, init=True)
                post_init(self, *args, **kwargs)  # pyright: ignore
                if self.__class__ is _make_initable_wrapper(cls):
                    # Convert all the fields filled in by `__post_init__` as well.
                    _convert_fields(self, init=False)

            cls.__post_init__ = __post_init__  # pyright: ignore
        else:
            post_init = None

        # Fairly common to write `Superclass.__init__.__doc__ = "..."` with
        # dataclass-provided inits; here we look through the class hierarchy and will
        # copy this doc forward.
        if has_dataclass_init:
            init_doc = cls.__init__.__doc__

        # [Step 4] Register as a dataclass.
        cls = dataclass(eq=False, repr=False, frozen=True, init=has_dataclass_init)(
            cls  # pyright: ignore
        )
        # [Step 3b] -- finish off building `__init__` methods. Until we'd done
        # dataclass'ification then we didn't necessarily have our `__init__` method.

        # Set annotation to the converter input. This is useful for runtime type
        # checkers.
        # Note that mutating the `__init__.__annotations__` is okay, as it was created
        # by the dataclass decorator on the previous line, so nothing else owns it.
        for f in dataclasses.fields(cls):
            if f.name not in cls.__init__.__annotations__:
                continue  # Odd behaviour, so skip.
            try:
                converter = f.metadata["converter"]
            except KeyError:
                pass
            else:
                try:
                    signature = inspect.signature(converter)
                except ValueError:
                    # e.g. `inspect.signature(str)` fails
                    converter_annotation = Any
                else:
                    parameters = list(signature.parameters.values())
                    if len(parameters) == 0:
                        # No idea what happened, but play it safe.
                        converter_annotation = Any
                    else:
                        converter_annotation = parameters[0].annotation
                        if converter_annotation is inspect.Signature.empty:
                            converter_annotation = Any
                cls.__init__.__annotations__[f.name] = converter_annotation

        # Registering here records that the `dataclass(...)` call has happened.
        _has_dataclass_init[cls] = has_dataclass_init

        # Now handle conversion for cases (a) and (b) above, in which there is no
        # `__post_init__`.
        if post_init is None:
            init = cls.__init__

            @ft.wraps(init)
            def __init__(self, *args, **kwargs):
                __tracebackhide__ = True
                init(self, *args, **kwargs)
                # Same `if` trick as with `__post_init__`.
                if self.__class__ is _make_initable_wrapper(cls):
                    _convert_fields(self, init=True)
                    _convert_fields(self, init=False)

            cls.__init__ = __init__

        # Assign `__doc__` in case it has been manually overridden:
        # ```
        # class Foo(eqx.Module):
        #     x: int
        #
        # Foo.__init__.__doc__ = "Foo should be called with with an integer `x`."
        #
        # class Bar(Foo):
        #     pass
        #
        # # Now we try to access `Bar.__init__.__doc__`. (E.g. during docgen.)
        # ```
        if has_dataclass_init:
            cls.__init__.__doc__ = init_doc  # pyright: ignore
            # TODO: is this next line still necessary?
            cls.__init__.__module__ = cls.__module__

        # [Step 5] We support an optional `strict` mode for Rust-like strictness in the
        # type checking.
        # In practice this is probably too much for your average user, but it's a great
        # way to build robust libraries.
        _is_force_abstract[cls] = strict_config.force_abstract
        _is_strict[cls] = strict
        if strict:
            for base in bases:
                if base is Module:
                    # We shouldn't check `Module` itself: we can't make it strict as to
                    # be able to inherit from it we would need to make it abstract. To
                    # be able to make it abstract its name must start with `Abstract`.
                    # That we cannot do for backward compatibility.
                    continue
                if _is_special_form(base):
                    # Skip `typing.Generic` etc.
                    continue
                # Invariant: all base classes are also strict modules.
                if not issubclass(base, Module):
                    raise TypeError(
                        "Strict `eqx.Module`s must only inherit from other strict "
                        f"`eqx.Module`s. `{cls.__module__}.{cls.__qualname__}` is a "
                        "strict Module inheriting from "
                        f"`{base.__module__}.{base.__qualname__}`, which is not a "
                        "`Module` at all."
                    )
                if not _is_strict[base]:
                    raise TypeError(
                        "Strict `eqx.Module`s must only inherit from other strict "
                        f"`eqx.Module`s. `{cls.__module__}.{cls.__qualname__}` is a "
                        "strict Module inheriting from "
                        f"`{base.__module__}.{base.__qualname__}`, which is a "
                        "`Module`, but not a strict `Module`."
                    )
                # Invariant: concrete means final.
                if not _is_abstract(base):
                    raise TypeError(
                        "Every strict `eqx.Module` must be either abstract or final. "
                        "This means that it is not possible to inherit from a concrete "
                        f"strict `eqx.Module`. `{cls.__module__}.{cls.__qualname__}` "
                        "is a strict Module inheriting from "
                        f"`{base.__module__}.{base.__qualname__}`, which is a concrete "
                        "strict `Module`."
                    )
                # Invariant: field definitions and __init__ methods must all appear on
                # just one class in a hierarchy.
                base_num_fields = len(dataclasses.fields(base))
                if (base_num_fields > 0) or (not _has_dataclass_init[base]):
                    # If we've added more fields, or added a custom init method, then
                    # error.
                    if len(dataclasses.fields(cls)) != base_num_fields:
                        raise TypeError(
                            "For readability, any custom `__init__` method, and all "
                            "fields, must all be defined on the same strict Module. "
                            f"{cls.__module__}.{cls.__qualname__} is a strict Module "
                            "that is attempting to add fields, and inherit from "
                            f"{base.__module__}.{base.__qualname__}. However the "
                            "latter already has fields or a custom `__init__` defined."
                        )
                    if added_custom_init:
                        raise TypeError(
                            "For readability, any custom `__init__` method, and all "
                            "fields, must all be defined on the same strict Module. "
                            f"{cls.__module__}.{cls.__qualname__} is a strict Module "
                            "that is attempting to define a custom `__init__` method, "
                            f"and inherit from {base.__module__}.{base.__qualname__}. "
                            "However the latter already has fields "
                            "or a custom `__init__` defined."
                        )
            if not strict_config.allow_abstract_name:
                has_abstract_name = cls.__name__.startswith(
                    "Abstract"
                ) or cls.__name__.startswith("_Abstract")
                if _is_abstract(cls):
                    if not has_abstract_name:
                        # Invariant: abstract classes have names beginning with
                        # `Abstract`.
                        main = (
                            "Abstract strict `eqx.Module`s must be named starting "
                            f"with 'Abstract' or '_Abstract'. Got {name} when defining "
                            f"{cls.__module__}.{cls.__qualname__}."
                        )
                        if _is_force_abstract[cls]:
                            raise TypeError(main)
                        inner = []
                        if len(cls.__abstractmethods__) > 0:
                            inner.append(
                                f"abstract methods: {list(cls.__abstractmethods__)}"
                            )
                        if len(cls.__abstractvars__) > 0:
                            inner.append(
                                f"abstract variables: {list(cls.__abstractvars__)}"
                            )
                        if len(cls.__abstractclassvars__) > 0:
                            inner.append(
                                (
                                    "abstract class variables: "
                                    f"{list(cls.__abstractclassvars__)}"
                                )
                            )
                        inner = ", ".join(inner)
                        inner = " " + inner + "."
                        raise TypeError(main + inner)
                else:
                    if has_abstract_name:
                        # Invariant: concrete classes do not have names beginning with
                        # `Abstract`.
                        raise TypeError(
                            "Concrete strict `eqx.Module`s should not have names "
                            f"starting with 'Abstract' or '_Abstract'. Got '{name}' "
                            f"when defining {cls.__module__}.{cls.__qualname__}.",
                        )
            for k, v in cls.__dict__.items():
                if isinstance(v, _wrap_method):
                    v = v.method
                    # Invariant: concrete methods are not overridden.
                    if not getattr(v, "__isabstractmethod__", False):
                        for base in bases:
                            old_v = getattr(base, k, _dummy_abstract)
                            if not inspect.isfunction(old_v):
                                raise TypeError(
                                    "Strict `eqx.Module`s cannot override non-methods "
                                    "with methods. "
                                    f"`{cls.__module__}.{cls.__qualname__}.{k}` is "
                                    "attempting to override "
                                    f"`{base.__module__}.{base.__qualname__}.{k}`."
                                )
                            if not strict_config.allow_method_override and not getattr(
                                old_v, "__isabstractmethod__", False
                            ):
                                raise TypeError(
                                    "Strict `eqx.Module`s cannot override concrete "
                                    "methods. "
                                    f"`{cls.__module__}.{cls.__qualname__}.{k}` is "
                                    "attempting to override "
                                    f"`{base.__module__}.{base.__qualname__}.{k}`."
                                )
        # [Step 6] Register as a pytree.
        jtu.register_pytree_with_keys(
            cls,
            flatten_with_keys=ft.partial(_flatten_module, with_keys=True),  # pyright: ignore
            flatten_func=ft.partial(_flatten_module, with_keys=False),  # pyright: ignore
            unflatten_func=ft.partial(_unflatten_module, cls),  # pyright: ignore
        )
        # Done!
        return cls

    @property
    def __signature__(cls):
        # Use signature of __init__ method for non-callable equinox modules
        sig = inspect.signature(cls.__init__)
        params = list(sig.parameters.values())[1:]  # Remove self parameter
        return sig.replace(parameters=params)

    # This method is called whenever you initialise a module: `MyModule(...)`
    def __call__(cls, *args, **kwargs):
        __tracebackhide__ = True
        if _is_force_abstract[cls]:
            # Any other is-abstract checks will be handled in super().__call__.
            raise TypeError("Cannot instantiate abstract `equinox.Module`.")
        if _has_dataclass_init[cls]:
            for x in jtu.tree_leaves((args, kwargs)):
                _warn_jax_transformed_function(cls, x)
            # else it's handled in __setattr__, but that isn't called here.
        # [Step 1] Modules are immutable -- except during construction. So defreeze
        # before init.
        initable_cls = _make_initable_wrapper(cls)
        # [Step 2] Instantiate the class as normal.
        self = super(_ActualModuleMeta, initable_cls).__call__(*args, **kwargs)
        assert not _is_abstract(cls)
        # [Step 3] Check that all fields are occupied.
        missing_names = {
            field.name
            for field in dataclasses.fields(cls)  # pyright: ignore
            # Not `vars` or `__dict__`, to allow for `property`s overwriting a field.
            # Not recommended, but allowable for backward compatibility.
            if field.name not in dir(self)
        }
        if len(missing_names):
            raise ValueError(
                f"The following fields were not initialised during __init__: "
                f"{missing_names}"
            )
        # [Step 3.5] Prevent arrays from being marked as static
        for field in dataclasses.fields(self):
            if field.metadata.get("static", False):
                if any(
                    jtu.tree_map(
                        is_array, jtu.tree_flatten(getattr(self, field.name))[0]
                    )
                ):
                    warnings.warn(
                        "A JAX array is being set as static! This can result "
                        "in unexpected behavior and is usually a mistake to do.",
                        stacklevel=2,
                    )
        # Freeze.
        object.__setattr__(self, "__class__", cls)
        # [Step 4] Run any custom validators. (After freezing; as they run
        # unconditionally across the whole MRO, they aren't allowed to mutate.)
        for kls in cls.__mro__:
            try:
                check = kls.__dict__["__check_init__"]
            except KeyError:
                pass
            else:
                check(self)
        return self

    # This bit is kind of sneaky. This is the reason that `_has_dataclass_init` is done
    # after dataclass registration -- we sneakily treat it as a marker for whether
    # dataclass registration has happened yet. And if it hasn't, we hide any
    # `__wrapped__` attribute. We want such attributes for the sake of
    # `module_update_wrapper`, but if `dataclass` sees it then it tries to follow it.
    def __getattribute__(cls, item):
        value = super().__getattribute__(item)
        if (
            item == "__wrapped__"
            and isinstance(value, property)
            and cls not in _has_dataclass_init
        ):
            raise AttributeError
        else:
            return value

    def __setattr__(cls, item, value):
        if _not_magic(item) and inspect.isfunction(value):
            value = _wrap_method(value)
        super().__setattr__(item, value)


if TYPE_CHECKING:

    @dataclass_transform(field_specifiers=(dataclasses.field, field, static_field))
    class _ModuleMeta(abc.ABCMeta):
        __abstractvars__: frozenset[str]
        __abstractclassvars__: frozenset[str]
else:
    _ModuleMeta = _ActualModuleMeta


def _is_special_form(cls):
    # This function is basically a heuristic hack.
    # If you're getting spurious warnings from this, and think you have another kind of
    # class that should be excluded from Equinox's checks, then please open a GitHub
    # issue: https://github.com/patrick-kidger/equinox/issues
    if cls is _Initable:
        return True
    if cls.__module__ in ("typing", "typing_extensions", "collections.abc"):
        return True
    if Protocol in cls.__bases__:
        return True
    return False


def _not_magic(k: str) -> bool:
    return not (k.startswith("__") and k.endswith("__"))


class _wrap_method:
    def __init__(self, method):
        self.method = method
        if getattr(self.method, "__isabstractmethod__", False):
            self.__isabstractmethod__ = self.method.__isabstractmethod__

    def __get__(self, instance, owner):
        if instance is None:
            return self.method
        else:
            # Why `inplace=True`?
            # This is safe because the `BoundMethod` was only instantiated here.
            # This is necessary so that `_method.__self__ is instance`, which is used
            # as part of a no-cycle check in `_make_initable`.
            _method = _module_update_wrapper(
                BoundMethod(self.method, instance), None, inplace=True
            )
            return _method


_dummy_abstract = abc.abstractmethod(lambda self: 1)
_is_force_abstract = weakref.WeakKeyDictionary()
_is_strict = weakref.WeakKeyDictionary()
_has_dataclass_init = weakref.WeakKeyDictionary()


def _is_abstract(cls):
    return (
        _is_force_abstract[cls]
        or len(cls.__abstractmethods__) > 0
        or len(cls.__abstractvars__) > 0
        or len(cls.__abstractclassvars__) > 0
    )


_wrapper_field_names = {
    "__module__",
    "__name__",
    "__qualname__",
    "__doc__",
    "__annotations__",
}


class _Initable:
    # Prevent `__init_subclass__` from triggering when creating initable versions of
    # classes.
    def __init_subclass__(cls, **kwargs):
        del kwargs


_transform_types = {
    type(transform(lambda x: x))
    for transform in (
        jax.jit,
        jax.grad,
        jax.vmap,
        jax.value_and_grad,
        jax.jacfwd,
        jax.jacrev,
        jax.hessian,
        jax.custom_jvp,
        jax.custom_vjp,
        jax.checkpoint,  # pyright: ignore
        jax.pmap,
    )
}


def _warn_jax_transformed_function(cls, x):
    # not `isinstance`, just in case JAX every tries to override `__instancecheck__`.
    if type(x) in _transform_types:

        class _JaxTransformException(Exception):
            pass

        def _is_array_like(x):
            if is_array_like(x):
                raise _JaxTransformException

        while True:
            try:
                x = x.__wrapped__
            except AttributeError:
                break
            try:
                jtu.tree_map(_is_array_like, x)
            except _JaxTransformException:
                warnings.warn(
                    f"""
Possibly assigning a JAX-transformed callable as an attribute on
{cls.__module__}.{cls.__qualname__}. This will not have any of its parameters updated.

For example, the following code is buggy:
```python
class MyModule(eqx.Module):
vmap_linear: Callable

def __init__(self, ...):
    self.vmap_linear = jax.vmap(eqx.nn.Linear(...))

def __call__(self, ...):
    ... = self.vmap_linear(...)
```
This is because the callable returned from `jax.vmap` is *not* a PyTree. This means that
the parameters inside the `eqx.nn.Linear` layer will not receive gradient updates.

You can most easily fix this either by applying the wrapper at `__call__` time:
```python
class MyModule(eqx.Module):
linear: Callable

def __init__(self, ...):
    self.linear = eqx.nn.Linear(...)

def __call__(self, ...):
    ... = jax.vmap(self.linear)(...)
```
or by using `eqx.filter_vmap` instead (which *does* return a PyTree):
```python
class MyModule(eqx.Module):
vmap_linear: Callable

def __init__(self, ...):
    self.vmap_linear = eqx.filter_vmap(eqx.nn.Linear(...))

def __call__(self, ...):
    ... = self.vmap_linear(...)
```
""",
                    stacklevel=3,
                )
                break


def _make_initable_wrapper(cls: _ActualModuleMeta) -> _ActualModuleMeta:
    post_init = getattr(cls, "__post_init__", None)
    return _make_initable(cls, cls.__init__, post_init, wraps=False)


@ft.lru_cache(maxsize=128)
def _make_initable(
    cls: _ActualModuleMeta, init, post_init, wraps: bool
) -> _ActualModuleMeta:
    # Used as part of the key. Don't cache if these have changed.
    # In practice, monkey-patching these on the class -- after you've already
    # instantiated it somewhere! -- is an *ahem*, adventurous, thing to do. But never
    # let it be said that Equinox doesn't support you in your questionable life choices!
    del init, post_init

    if wraps:
        field_names = _wrapper_field_names
    else:
        field_names = {field.name for field in dataclasses.fields(cls)}

    class _InitableModule(_Initable, cls):
        pass

    def __setattr__(self, name, value):
        if name in field_names:
            if isinstance(value, BoundMethod) and value.__self__ is self:
                raise ValueError(
                    """Cannot assign methods in __init__.

That is, something like the following is not allowed:
```
class MyModule(eqx.Module):
    foo: Callable

    def __init__(self):
        self.foo = self.bar

    def bar(self):
        ...
```
this is because Equinox modules are PyTrees -- but the above does not have a tree
structure! `self.foo.__self__` is a cycle that brings us back to `self`.

In the above example, you probably want something like this instead:
```
class MyModule(eqx.Module):
    @property
    def foo(self):
        return self.bar

    def bar(self):
        ...
```
so that you can still use `self.foo`, but it is not stored in the PyTree structure.

This is a check that was introduced in Equinox v0.11.0. Before this, the above error
went uncaught, possibly leading to silently wrong behaviour.
"""
                )
            else:
                _warn_jax_transformed_function(type(self), value)
                object.__setattr__(self, name, value)
        else:
            raise AttributeError(f"Cannot set attribute {name}")

    # Done like this to avoid dataclasses complaining about overriding setattr on a
    # frozen class.
    _InitableModule.__setattr__ = __setattr__
    # Make beartype happy
    _InitableModule.__init__ = cls.__init__  # pyright: ignore
    # Appears in AbstractVar error messages
    _InitableModule.__name__ = cls.__name__
    _InitableModule.__qualname__ = cls.__qualname__
    # I don't have a specific use-case for this but it's probably good practice.
    _InitableModule.__module__ = cls.__module__

    return _InitableModule


cache_clears.append(_make_initable.cache_clear)


def _convert_fields(module, init: bool):
    for field in dataclasses.fields(module):
        if field.init is init:
            try:
                converter = field.metadata["converter"]
            except KeyError:
                pass
            else:
                try:
                    value = getattr(module, field.name)
                except AttributeError:
                    # Let the all-fields-are-filled check handle the error.
                    pass
                else:
                    setattr(module, field.name, converter(value))


# Used to provide a pretty repr when doing `jtu.tree_structure(SomeModule(...))`.
@dataclass()
class _FlattenedData:
    dynamic_field_names: tuple
    static_field_names: tuple
    static_field_values: tuple
    wrapper_field_names: tuple
    wrapper_field_values: tuple

    def __repr__(self):
        x = (
            self.dynamic_field_names,
            self.static_field_names,
            self.static_field_values,
        )
        return repr(x)[1:-1]


def _flatten_module(module: "Module", with_keys: bool):
    # Subnodes in the PyTree
    dynamic_field_names = []
    dynamic_field_values = []
    # Static metadata, placed in aux.
    static_field_names = []
    static_field_values = []
    # Python metadata like `__doc__` and `__module__`.
    wrapper_field_names = []
    wrapper_field_values = []

    for field_ in dataclasses.fields(module):
        name = field_.name
        try:
            # Not `getattr` so that we don't pick up `property`s.
            value = module.__dict__[name]
        except KeyError:
            # Uninitialised values during `__init__`, or when `property`s overwrite a
            # field.
            continue
        if field_.metadata.get("static", False):
            static_field_names.append(name)
            static_field_values.append(value)
        else:
            dynamic_field_names.append(name)
            if with_keys:
                dynamic_field_values.append((jtu.GetAttrKey(name), value))
            else:
                dynamic_field_values.append(value)
    sentinel = object()
    for name in _wrapper_field_names:
        value = getattr(module, name, sentinel)
        if value is not sentinel:
            wrapper_field_names.append(name)
            wrapper_field_values.append(value)
    aux = _FlattenedData(
        tuple(dynamic_field_names),
        tuple(static_field_names),
        tuple(static_field_values),
        tuple(wrapper_field_names),
        tuple(wrapper_field_values),
    )
    return tuple(dynamic_field_values), aux


def _unflatten_module(cls: type["Module"], aux: _FlattenedData, dynamic_field_values):
    # This doesn't go via `__init__`. A user may have done something nontrivial there,
    # and the field values may be dummy values as used in various places throughout JAX.
    # See also
    # https://jax.readthedocs.io/en/latest/pytrees.html#custom-pytrees-and-initialization,
    # which was (I believe) inspired by Equinox's approach here.
    module = object.__new__(cls)
    for name, value in zip(aux.dynamic_field_names, dynamic_field_values):
        object.__setattr__(module, name, value)
    for name, value in zip(aux.static_field_names, aux.static_field_values):
        object.__setattr__(module, name, value)
    for name, value in zip(aux.wrapper_field_names, aux.wrapper_field_values):
        object.__setattr__(module, name, value)
    return module


class Module(metaclass=_ModuleMeta):
    """Base class. Create your model by inheriting from this.

    This will make your model a
    [dataclass](https://docs.python.org/3/library/dataclasses.html) and a
    [pytree](https://jax.readthedocs.io/en/latest/pytrees.html).

    **Fields**

    Specify all its fields at the class level (identical to
    [dataclasses](https://docs.python.org/3/library/dataclasses.html)). This defines
    its children as a PyTree.

    ```python
    class MyModule(equinox.Module):
        weight: jax.Array
        bias: jax.Array
        submodule: equinox.Module
    ```

    **Initialisation**

    A default `__init__` is automatically provided, which just fills in fields with
    the arguments passed. For example `MyModule(weight, bias, submodule)`.

    Alternatively (quite commonly) you can provide an `__init__` method yourself:

    ```python
    class MyModule(equinox.Module):
        weight: jax.Array
        bias: jax.Array
        submodule: equinox.Module

        def __init__(self, in_size, out_size, key):
            wkey, bkey, skey = jax.random.split(key, 3)
            self.weight = jax.random.normal(wkey, (out_size, in_size))
            self.bias = jax.random.normal(bkey, (out_size,))
            self.submodule = equinox.nn.Linear(in_size, out_size, key=skey)
    ```

    **Methods**

    It is common to create some methods on the class -- for example to define the
    forward pass of a model.

    ```python
    class MyModule(equinox.Module):
        ...  # as above

        def __call__(self, x):
            return self.submodule(x) + self.weight @ x + self.bias
    ```

    !!! tip

        You don't have to define `__call__`:

        - You can define other methods if you want.
        - You can define multiple methods if you want.
        - You can define no methods if you want. (And just use `equinox.Module` as a
            nice syntax for custom PyTrees.)

        No method is special-cased.

    **Usage**

    After you have defined your model, then you can use it just like any other
    PyTree -- that just happens to have some methods attached. In particular you can
    pass it around across `jax.jit`, `jax.grad` etc. in exactly the way that you're
    used to.

    !!! example

        If you wanted to, then it would be completely safe to do

        ```python
        class MyModule(equinox.Module):
            ...

            @jax.jit
            def __call__(self, x):
                ...
        ```

        because `self` is just a PyTree. Unlike most other neural network libraries,
        you can mix Equinox and native JAX without any difficulties at all.

    !!! tip "For fans of strong typing."

        Equinox modules are all [ABCs](https://docs.python.org/3/library/abc.html)
        by default. This means you can use
        [`abc.abstractmethod`](https://docs.python.org/3/library/abc.html#abc.abstractmethod).
        You can also create abstract instance attributes or abstract class
        attributes, see [`equinox.AbstractVar`][] and
        [`equinox.AbstractClassVar`][].
    """  # noqa: E501

    #     [`equinox.AbstractClassVar`][]. Finally, some optional strict type-checking
    #     may be enabled by passing `strict=True`, e.g.
    #     `class Foo(eqx.Module, strict=True)`; see
    #     [strict modules](../advanced_fields/#strict-modules) for details.
    # """  # noqa: E501

    def __hash__(self):
        return hash(tuple(jtu.tree_leaves(self)))

    def __eq__(  # pyright: ignore
        self, other
    ) -> Union[bool, np.bool_, Bool[Array, ""]]:
        return tree_equal(self, other)

    def __repr__(self):
        return tree_pformat(self)


# Not using `jax.tree_util.Partial` as it doesn't implement __eq__ very well. See #480.
class BoundMethod(Module):
    """Just like a normal Python bound method... except that this one is a PyTree!

    This stores `__self__` as a subnode.
    """

    __func__: types.FunctionType = field(static=True)
    __self__: Module

    def __call__(self, *args, **kwargs):
        __tracebackhide__ = True
        return self.__func__(self.__self__, *args, **kwargs)

    @property
    def __wrapped__(self):
        return self.__func__.__get__(  # pyright: ignore
            self.__self__, type(self.__self__)
        )

    # This should be unnecessary in principle. In practice something goes wrong on
    # Python 3.9 and it returns the wrong thing.
    @property
    def __signature__(self):
        return inspect.signature(self.__wrapped__)


#
# Part 3: some downstream pieces. These don't actually affect the core `Module`
# abstraction, but this file is a convenient place to put them.
#


def module_update_wrapper(
    wrapper: Module, wrapped: Optional[Callable[_P, _T]] = None
) -> Callable[_P, _T]:
    """Like `functools.update_wrapper` (or its better-known cousin, `functools.wraps`),
    but acts on [`equinox.Module`][]s, and does not modify its input (it returns the
    updated module instead).

    !!! Example

        ```python
        class Wrapper(eqx.Module):
            fn: Callable

            def __call__(self, *args, **kwargs):
                return self.fn(*args, **kwargs)

            @property
            def __wrapped__(self):
                return self.fn

        def make_wrapper(fn):
            return eqx.module_update_wrapper(Wrapper(fn))
        ```

    For example, [`equinox.filter_jit`][] returns a module representing the JIT'd
    computation. `module_update_wrapper` is used on this module to indicate that this
    JIT'd computation wraps the original one. (Just like how `functools.wraps` is used.)

    Note that as in the above example, the wrapper class must supply a `__wrapped__`
    property, which redirects to the wrapped object.

    **Arguments:**

    - `wrapper`: the instance of the wrapper.
    - `wrapped`: optional, the callable that is being wrapped. If omitted then
        `wrapper.__wrapped__` will be used.

    **Returns:**

    A copy of `wrapper`, with the attributes `__module__`, `__name__`, `__qualname__`,
    `__doc__`, and `__annotations__` copied over from the wrapped function.
    """
    return _module_update_wrapper(wrapper, wrapped, inplace=False)


def _module_update_wrapper(
    wrapper: Module, wrapped: Optional[Callable[_P, _T]], inplace: bool
) -> Callable[_P, _T]:
    cls = wrapper.__class__
    if not isinstance(getattr(cls, "__wrapped__", None), property):
        raise ValueError("Wrapper module must supply `__wrapped__` as a property.")

    if wrapped is None:
        wrapped = wrapper.__wrapped__  # pyright: ignore

    if not inplace:
        # Make a clone, to avoid mutating the original input.
        leaves, treedef = jtu.tree_flatten(wrapper)
        wrapper = jtu.tree_unflatten(treedef, leaves)

    initable_cls = _make_initable(cls, None, None, wraps=True)
    object.__setattr__(wrapper, "__class__", initable_cls)
    try:
        # Like `ft.update_wrapper(wrapper, wrapped, updated=())`.
        # We don't update __dict__ as it's common/possible for wrapper and wrapped to
        # both be classes implementing __call__, in which case copying __dict__ over
        # basically just breaks the wrapper class.
        # We don't set __wrapped__, and instead demand that the wrapper class tell us
        # how to redirect to the wrapped object. This is avoid duplicating part of the
        # PyTree.
        for attr in _wrapper_field_names:
            try:
                value = getattr(wrapped, attr)
            except AttributeError:
                pass
            else:
                setattr(wrapper, attr, value)
    finally:
        object.__setattr__(wrapper, "__class__", cls)
    return cast(Callable[_P, _T], wrapper)


class Partial(Module):
    """Like `functools.partial`, but treats the wrapped function, and partially-applied
    value: Any = field(static=True)
    args and kwargs, as a PyTree.

    This is very much like `jax.tree_util.Partial`. The difference is that the JAX
    version requires that `func` be specifically a *function* -- and will silently
    misbehave if given any non-function callable, e.g. [`equinox.nn.MLP`][]. In contrast
    the Equinox version allows for arbitrary callables.
    """

    func: Callable
    args: tuple[Any, ...]
    keywords: dict[str, Any]

    def __init__(self, func, /, *args, **kwargs):
        """**Arguments:**

        - `func`: the callable to partially apply.
        - `*args`: any positional arguments to apply.
        - `**kwargs`: any keyword arguments to apply.
        """
        self.func = func
        self.args = args
        self.keywords = kwargs

    def __call__(self, *args, **kwargs):
        """Call the wrapped `self.func`.

        **Arguments:**

        - `*args`: the arguments to apply. Passed after those arguments passed during
            `__init__`.
        - `**kwargs`: any keyword arguments to apply.

        **Returns:**

        The result of the wrapped function.
        """
        return self.func(*self.args, *args, **kwargs, **self.keywords)


class Static(Module):
    """Wraps a value into a `eqx.field(static=True)`.

    This is useful to treat something as just static metadata with respect to a JAX
    transformation; for example this is used to return non-arrays from a filtered
    transform.
    """

    _leaves: list[Any] = field(static=True)
    _treedef: PyTreeDef = field(static=True)  # pyright: ignore

    def __init__(self, value: Any):
        # By flattening, we handle pytrees without `__eq__` methods.
        # When comparing static metadata for equality, this means we never actually
        # call `value.__eq__`.
        self._leaves, self._treedef = jtu.tree_flatten(value)

    @property
    def value(self):
        return jtu.tree_unflatten(self._treedef, self._leaves)



from .common import (
    buffer_at_set as buffer_at_set,
    maybe_set_p as maybe_set_p,
    MaybeBuffer as MaybeBuffer,
    select_if_vmap_p as select_if_vmap_p,
)
from .loop import scan as scan, while_loop as while_loop



from typing import Optional, Union

import jax
import jax.numpy as jnp
from jaxtyping import Array

from .._module import Module


class PReLU(Module, strict=True):
    """PReLU activation function.

    This is the elementwise function `x -> max(x, 0) + α * min(x, 0)`.
    This can be thought of as a leaky ReLU, with a learnt leak `α`.
    """

    negative_slope: Array

    def __init__(
        self,
        init_alpha: Optional[Union[float, Array]] = 0.25,
    ):
        r"""**Arguments:**

        - `init_alpha`: The initial value $\alpha$ of the negative slope.
            This should either be a `float` (default value is $0.25$), or
            a JAX array of $\alpha_i$ values. The shape of such a JAX array
            should be broadcastable to the input.
        """

        self.negative_slope = jnp.asarray(init_alpha)

    @jax.named_scope("eqx.nn.PReLU")
    def __call__(self, x: Array) -> Array:
        r"""**Arguments:**

        - `x`: The input.

        **Returns:**

        A JAX array of the same shape as the input.
        """
        with jax.numpy_dtype_promotion("standard"), jax.numpy_rank_promotion("allow"):
            return jnp.where(x >= 0, x, self.negative_slope * x)
